import numpy, os
import pytest

from py_exp_calc.exp_calc import *

__author__ = "seoanezonjic"
__copyright__ = "seoanezonjic"
__license__ = "MIT"

####################################
#DATA
####################################

@pytest.fixture
def first_list():
    return [1,2,3,4,5]

@pytest.fixture
def second_list():
    return [3,4,5,6,7]

@pytest.fixture
def matrix():
    return np.array([[3,  2,  1,   4], 
                     [5,  3,  7,   8], 
                     [9,  10, 3, 12],
                     [13, 14, 15, 3]])

@pytest.fixture
def net():
    return np.array([[0, 1, 1, 0, 0],
                     [1, 0, 1, 0, 1],
                     [1, 1, 0, 1, 0],
                     [0, 0, 1, 0, 1],
                     [0, 1, 0, 1, 0]])

@pytest.fixture
def mutation_like_branched_matrix():
    return np.array([[9, 9, 1, 1],
                     [9, 1, 1, 1],
                     [1, 1, 9, 9],
                     [1, 1, 1, 9],
                     [1, 1, 1, 1]])

@pytest.fixture
def rectangular_matrix():
    return np.array([[1,  1,  0], 
                     [0,  0,  1], 
                     [0,  0,  1],
                     [1,  1,  0],
                     [1,  1,  1]])

@pytest.fixture
def rowIds():
    return ['P1', 'P2', 'P3', 'P4', 'P5']

@pytest.fixture
def colIds():
    return ['D1', 'D2', 'D3']



####################################
# LIST OPERATIONS
####################################

# One List operation
####################################

def test_get_stats_from_list():
    sample = [0, 0, 1, 1, 2, 2, 3, 3, 4, 4]
    
    expected = [['Elements', '10'], ['Elements Non Zero', '8'], ['Non Zero Density', '0.8'], ['Max', '4'],
                ['Min', '0'], ['Average', '2.0'], ['Variance', '2.0'], ['Standard Deviation', '1.4142135623730951'],
                ['Q1', '1.0'], ['Median', '2.0'], ['Q3', '3.0'], ['Min Non Zero', '1'], ['Average Non Zero', '2.5'],
                ['Variance Non Zero', '1.25'], ['Standard Deviation Non Zero', '1.118033988749895'], ['Q1 Non Zero', '1.75'],
                ['Median Non Zero', '2.5'], ['Q3 Non Zero', '3.25']]
    
    returned = list(get_stats_from_list(sample))
    assert expected == returned

def test_uniq():
    arr1 = [1,1,1,2,3,3,2,1]
    test_result = uniq(arr1)
    expected_result = [1,2,3]
    assert test_result == expected_result

def test_flatten():
    chaos_list = [["a", ["b", ["c"]]], ["d"], [[["e"]]], ["f", [["g"], ["h"], [[["i"]]]]],[[[]]],["j","q",["l","m"],["n"],[[["o"]]]],[[[[[[[[[[["Un pato cuantico"]]]]]]]]]]]]
    expected = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j','q','l','m','n','o', "Un pato cuantico"]
    returned = flatten(chaos_list)
    assert expected == returned

# List x List operation
####################################

def test_intersection(first_list, second_list):
    assert intersection(first_list, second_list) == [3,4,5]
    assert intersection(first_list, second_list, indexing=True) == [3,4,5]

def test_union(first_list, second_list):
    assert union(first_list, second_list) == [1,2,3,4,5,6,7]
    assert union(first_list, second_list, indexing=True) == [1,2,3,4,5,6,7]
    
def test_diff(first_list, second_list):
    assert diff(first_list, second_list) == [1,2]
    assert diff(first_list, second_list, indexing=True) == [1,2]

    assert diff(second_list, first_list) == [6,7]
    assert diff(second_list, first_list, indexing=True) == [6,7]

####################################
# DIC OPERATIONS
####################################

# Values - operation
####################################

## Get
######

def test_dig():
    nested_dict = {5 : { 7: 'a'}}
    assert dig(nested_dict, 5, 7) == 'a'
    assert dig(nested_dict, 5, 2) == None


def test_get_hash_values_idx():
    input_hash = {"P1": ["D4", "D5", "D3"], "P2": ["D3"], "P3": ["D5", "D8"]}
    expected = {"D4": 0, "D5": 1, "D3": 2, "D8": 3}
    returned = get_hash_values_idx(input_hash)
    assert expected == returned

# Insert
########

def test_add_record():
    dict_test = {}
    add_record(dict_test, "D1", "P1")
    assert dict_test == {"D1": ["P1"]}    
    add_record(dict_test, "D1", "P2", uniq=True)
    assert dict_test == {"D1": ["P1", "P2"]}    
    add_record(dict_test, "D1", "P2", uniq=True)
    assert dict_test == {"D1": ["P1", "P2"]}    
    add_record(dict_test, "D1", "P2")
    assert dict_test == {"D1": ["P1", "P2", "P2"]}   
    add_record(dict_test, "D2", "P3", uniq=True)
    assert dict_test == {"D1": ["P1", "P2", "P2"], "D2": ["P3"]}

def test_add_nested_value():
    input_hash = {}
    add_nested_value(input_hash, ("D1", "P1"), 3)
    assert input_hash == {"D1": {"P1": 3}}
    add_nested_value(input_hash, ("D1", "P2"), 5)
    assert input_hash == {"D1": {"P1": 3, "P2": 5}}
    add_nested_value(input_hash, ("D2", "P5"), 9)
    assert input_hash == {"D1": {"P1": 3, "P2": 5}, "D2": {"P5": 9}}

# Update
########

def test_update_innermost():
    default = {'legend': {'font': {'family': 'Courier', 'size': 12, 'weight': 'bold', 'color': 'black'}}, 'bordercolor': 'Black', 'borderwidth': 2}
    user = {'legend': {'font': {'size': 24}}, 'bordercolor': 'White'}

    #testing default dictionary update
    default_copy = default.copy()
    default_copy.update(user)
    assert default_copy == {'legend': {'font': {'size': 24}}, 'bordercolor': 'White', 'borderwidth': 2}

    #testing update_innermost function
    default_copy2 = default.copy()
    update_innermost(default_copy2, user)
    assert default_copy2 == {'legend': {'font': {'family': 'Courier', 'size': 24, 'weight': 'bold', 'color': 'black'}}, 'bordercolor': 'White', 'borderwidth': 2}

    #Testing that update_innermost does add nested dictionaries if key does not exist in default
    default_copy3 = default.copy()
    user = {'legend': {'font': {'size': 24}}, 'userUniqueKey': {'key1': 'value1', 'key2': 'value2'}}    
    update_innermost(default_copy3, user)
    assert default_copy3 == {'legend': {'font': {'family': 'Courier', 'size': 24, 'weight': 'bold', 'color': 'black'}}, 'bordercolor': 'Black', 'borderwidth': 2, 
                             'userUniqueKey': {'key1': 'value1', 'key2': 'value2'}}

# Keys - operation
####################################

def test_transform_keys():
    input_hash = {"A": 1, "B": 2, "C": 3}
    duplicate_string = lambda x: x + x
    
    expected = {"AA": 1, "BB": 2, "CC": 3}
    returned = transform_keys(input_hash, duplicate_string)
    assert expected == returned

def test_invert_hash():
    initial_hash = {"D1": ["P1", "P2"], "D2": ["P3"], "D3": ["P1"]}
    inverted_hash = {"P1": ["D1","D3"], "P2": ["D1"], "P3": ["D2"]}
    returned_inverted_hash = invert_hash(initial_hash)
    assert inverted_hash == returned_inverted_hash

def test_invert_nested_hash():
    initial_hash = {"D1": {"P1": 3, "P2": 4}, "D2": {"P3": 5}, "D3": {"P1": 9}}
    inverted_hash = {"P1": {"D1": 3, "D3": 9}, "P2": {"D1": 4}, "P3": {"D2": 5}}
    returned_inverted_hash = invert_nested_hash(initial_hash)
    assert inverted_hash == returned_inverted_hash

def test_remove_nested_entries():
    initial_hash = {"D1": {"P1": 3, "P2": 8}, "D2": {"P3": 2}, "D3": {"P1": 9}}
    filter_func = lambda k, v: v > 5

    expected = {"D1": {"P2": 8}, "D3": {"P1": 9}}
    
    remove_nested_entries(initial_hash, filter_func)
    assert expected == initial_hash

####################################
# ARRAY (Matrix) OPERATIONS
####################################

# I/O operations
####################################

def test_save_and_load():
    matrix = np.array([[1, 0, 0, 0], [1, 1, 0, 0], [0, 0, 1, 0], [1, 0, 0, 1]])
    x_axis_names = ['S1', 'S2', 'S3', 'S4']
    y_axis_names = ['S1', 'S2', 'S3', 'S4']
    
    save(matrix, 'test_matrix.npy', x_axis_names, 'test_x_axis.txt', y_axis_names, 'test_y_axis.txt')
    loaded_matrix, loaded_x_axis, loaded_y_axis = load('test_matrix.npy', 'test_x_axis.txt', 'test_y_axis.txt')
    
    assert (matrix == loaded_matrix).all()
    assert x_axis_names == loaded_x_axis
    assert y_axis_names == loaded_y_axis

    save(matrix, 'test_matrix.npy')
    loaded_matrix, loaded_x_axis, loaded_y_axis = load('test_matrix.npy')
    assert (matrix == loaded_matrix).all()
    assert None == loaded_x_axis
    assert None == loaded_y_axis

    #Deleting files
    os.remove('test_matrix.npy')
    os.remove('test_x_axis.txt')
    os.remove('test_y_axis.txt')

# Normalization
####################################

def test_normalize_matrix(matrix):
    expected_min_max = (matrix - 1) / 14.0 #min = 1, max = 15 in the input matrix
    returned_min_max = normalize_matrix(matrix, "min_max")
    assert (expected_min_max == returned_min_max).all()

    expected_cosine = matrix / 3.0 #all the values in the input matrix diagonal are 3, so square root of denominator (matrix[i, i] * matrix[j,j]) is always 3
    returned_cosine = normalize_matrix(matrix, "cosine")
    assert (expected_cosine == returned_cosine).all()

    expected_row_col = np.array([[0.3       , 0.13187609, 0.05423261, 0.18856181],
                                [0.32969024, 0.13043478, 0.25031949, 0.24866795],
                                [0.48809353, 0.35759927, 0.08823529, 0.306786  ],
                                [0.61282588, 0.43516891, 0.38348249, 0.06666667]])
    returned_row_col = normalize_matrix(matrix, "rows_cols")
    assert np.isclose(expected_row_col, returned_row_col).all()

    with pytest.raises(ValueError):
        normalize_matrix(matrix, "invalid_method")


def test_min_max_normalization_matrix(matrix):
    expected_min_max = (matrix - 1) / 14.0 #min = 1, max = 15 in the input matrix
    returned_min_max = min_max_normalization_matrix(matrix)
    assert (expected_min_max == returned_min_max).all()

def test_cosine_normalization_matrix(matrix):
    expected_cosine = matrix / 3.0 #all the values in the input matrix diagonal are 3, so square root of denominator (matrix[i, i] * matrix[j,j]) is always 3
    returned = cosine_normalization(matrix)
    assert (expected_cosine == returned).all()

def test_rows_cols_normalization_matrix(matrix):
    expected_row_col = np.array([[0.3       , 0.13187609, 0.05423261, 0.18856181],
                                [0.32969024, 0.13043478, 0.25031949, 0.24866795],
                                [0.48809353, 0.35759927, 0.08823529, 0.306786  ],
                                [0.61282588, 0.43516891, 0.38348249, 0.06666667]])
    returned_row_col = row_col_normalization(matrix)
    assert np.isclose(expected_row_col, returned_row_col).all()

# Sim <-> Dis conversion
####################################

def test_coords2sim(mutation_like_branched_matrix):
    #Expected behaviour: Row 5 dirverges to row 4 and row 2, row 4 diverges to row 3, row 2 diverges to row 1
    #Similary should be consistent with that
    dotProduct_expected =  [[164,  92,  36,  28,  20],
                            [ 92,  84,  28,  20,  12],
                            [ 36,  28, 164,  92,  20],
                            [ 28,  20,  92,  84,  12],
                            [ 20,  12,  20,  12,   4]]
    dotProduct_returned = coords2sim(mutation_like_branched_matrix)

    normalizedScaling_expected =   [[1.         , 0.5        , 0.         , 0.1339746 , 0.29289322],
                                    [0.5        , 1.         , 0.1339746  , 0.29289322, 0.5       ],
                                    [0.         , 0.1339746  , 1.         , 0.5       , 0.29289322],
                                    [0.1339746  , 0.29289322 , 0.5        , 1.        , 0.5       ],
                                    [0.29289322 , 0.5        , 0.29289322 , 0.5       , 1.        ]]
    normalizedScaling_returned = coords2sim(mutation_like_branched_matrix, "normalizedScaling")
    
    euclidean_expected =   [[1.        , 0.11111111, 0.05882353, 0.06731103, 0.0812103 ],
                            [0.11111111, 1.        , 0.06731103, 0.0812103 , 0.11111111],
                            [0.05882353, 0.06731103, 1.        , 0.11111111, 0.0812103 ],
                            [0.06731103, 0.0812103 , 0.11111111, 1.        , 0.11111111],
                            [0.0812103 , 0.11111111, 0.0812103 , 0.11111111, 1.        ]]
    euclidean_returned = coords2sim(mutation_like_branched_matrix, "euclidean")

    assert (dotProduct_expected == dotProduct_returned).all()
    assert np.isclose(normalizedScaling_expected, normalizedScaling_returned).all()
    assert np.isclose(euclidean_expected, euclidean_returned).all()
    with pytest.raises(ValueError):
        coords2sim(mutation_like_branched_matrix, sim="invalid_method")

def test_coords2dis(mutation_like_branched_matrix):
    #Expected behaviour: Row 5 dirverges to row 4 and row 2, row 4 diverges to row 3, row 2 diverges to row 1
    #Distances should be consistent with that
    expected = [[ 0.         ,  8.         , 16.         , 13.85640646 , 11.3137085 ],
                [ 8.         ,  0.         , 13.85640646 , 11.3137085  ,  8.        ],
                [16.         , 13.85640646 ,  0.         ,  8.         , 11.3137085 ],
                [13.85640646 , 11.3137085  ,  8.         ,  0.         ,  8.        ],
                [11.3137085  ,  8.         , 11.3137085  ,  8.         ,  0.        ]]
    returned = coords2dis(mutation_like_branched_matrix)
    assert np.isclose(expected, returned).all()

# Stats
####################################

def test_get_test():
    table_pair = [["F1","F2","V"],[0,1,1],[0,1,2],[0,1,3],[1,0,0],[1,0,0],[1,0,1]]
    result = get_test(table_pair, idx_factor=[0,1], header=True)
    expected = [['F1', 'V', 0.08244173716174748], ['F2', 'V', 0.08244173716174748]]
    assert result == expected

    table_pair = [[0,1,1],[0,1,2],[0,1,3],[1,0,0],[1,0,0],[1,0,1]]
    result = get_test(table_pair, idx_factor=[0,1], header=False)
    expected = [['F1', 'V1', 0.08244173716174748], ['F2', 'V1', 0.08244173716174748]]
    assert result == expected

    table_pair = [[0,1],[0,2],[1,3],[1,0],[2,0],[2,1]]
    result = get_test(table_pair, idx_factor=[0], header=False)
    expected = [['F1_0:1', 'V1', 1.0], ['F1_0:2', 'V1', 1.0], ['F1_1:2', 'V1', 1.0]]
    assert result == expected

def test_get_corr():
    vector1 = [1,2,3,4,5,6,7,8,9,10]
    vector2 = [10,9,8,7,6,5,4,3,2,1]
    vector3 = [2,4,6,8,10,12,14,16,18,20]

    np.random.seed(1)
    random_vect_1 = np.random.randint(-10000,10000, 10000)
    random_vect_2 = np.random.randint(-10000,10000, 10000)

    assert get_corr(vector1, vector2, corr_type= "pearson") == (-1.0, 0.0) #(Corr, P-value)
    assert get_corr(vector1, vector3, corr_type= "pearson") == (1.0, 0.0)
    random_cor, random_prop = get_corr(random_vect_1, random_vect_2, corr_type= "pearson")
    assert (round(random_cor, 4), round(random_prop, 4)) == (round(0.006686913147369477,4), round(0.5037409752662002,4))

    assert np.isclose(get_corr(vector1, vector2, corr_type="spearman"), (-1.0, 0.0)).all() #(Corr, P-value)
    assert np.isclose(get_corr(vector1, vector3, corr_type="spearman"), (1.0, 0.0)).all()
    random_cor, random_prop = get_corr(random_vect_1, random_vect_2, corr_type="spearman")
 
    assert np.isclose((random_cor, random_prop), (0.006617925733547477, 0.5081525499888688)).all()

    with pytest.raises(ValueError):
        get_corr(vector1, vector2, corr_type="invalid_method")

def test_pearsonr():
    vector1 = [1,2,3,4,5,6,7,8,9,10]
    vector2 = [10,9,8,7,6,5,4,3,2,1]
    vector3 = [2,4,6,8,10,12,14,16,18,20]

    np.random.seed(1)
    random_vect_1 = np.random.randint(-10000,10000, 10000)
    random_vect_2 = np.random.randint(-10000,10000, 10000)

    #Asserting correcteness of pearsonr using two vectors
    assert pearsonr(vector1, vector2) == (-1.0, 0.0) #(Corr, P-value)
    assert pearsonr(vector1, vector3) == (1.0, 0.0)    
    random_cor, random_prop = pearsonr(random_vect_1, random_vect_2)
    assert  (round(random_cor, 4), round(random_prop, 4)) == (round(0.006686913147369477, 4), round(0.5037409752662002, 4))

    #Asserting correcteness of pearsonr using a matrix with N columns
    matrix_like = np.array([vector1, vector2, vector3]).T
    cor_matrix, p_value_matrix = pearsonr(matrix_like)
    assert np.isclose(cor_matrix, np.array([[ 1., -1.,  1.], [-1.,  1., -1.], [ 1., -1.,  1.]])).all()
    assert np.isclose(p_value_matrix, np.array([[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]])).all()

    cor_matrix, p_value_matrix = pearsonr(matrix_like, alternative="greater")
    assert np.isclose(cor_matrix, np.array([[ 1., -1.,  1.], [-1.,  1., -1.], [ 1., -1.,  1.]])).all()
    assert np.isclose(p_value_matrix, np.array([[0., 1., 0.], [1., 0., 1.], [0., 1., 0.]])).all()

    cor_matrix, p_value_matrix = pearsonr(matrix_like, alternative="less")
    assert np.isclose(cor_matrix, np.array([[ 1., -1.,  1.], [-1.,  1., -1.], [ 1., -1.,  1.]])).all()
    assert np.isclose(p_value_matrix, np.array([[1., 0., 1.], [0., 1., 0.], [1., 0., 1.]])).all()

    with pytest.raises(ValueError):
        pearsonr(matrix_like, alternative="invalid_alternative")

def test_get_stats_from_matrix(net):
    expected = [['Matrix - Dimensions', '5x5'], ['Matrix - Elements', '25'], ['Matrix - Elements Non Zero', '12'], ['Matrix - Non Zero Density', '0.48'], ['Weigth - Max', '1'], ['Weigth - Min', '0'], ['Weigth - Average', '0.48'], ['Weigth - Variance', '0.24960000000000016'], ['Weigth - Standard Deviation', '0.499599839871872'], ['Weigth - Q1', '0.0'], ['Weigth - Median', '0.0'], ['Weigth - Q3', '1.0'], ['Weigth - Min Non Zero', '1'], ['Weigth - Average Non Zero', '1.0'], ['Weigth - Variance Non Zero', '0.0'], ['Weigth - Standard Deviation Non Zero', '0.0'], ['Weigth - Q1 Non Zero', '1.0'], ['Weigth - Median Non Zero', '1.0'], ['Weigth - Q3 Non Zero', '1.0'], ['Node - Elements', '5'], ['Node - Elements Non Zero', '5'], ['Node - Non Zero Density', '1.0'], ['Edges - Max', '3'], ['Edges - Min', '2'], ['Edges - Average', '2.4'], ['Edges - Variance', '0.24'], ['Edges - Standard Deviation', '0.4898979485566356'], ['Edges - Q1', '2.0'], ['Edges - Median', '2.0'], ['Edges - Q3', '3.0'], ['Edges - Min Non Zero', '2'], ['Edges - Average Non Zero', '2.4'], ['Edges - Variance Non Zero', '0.24'], ['Edges - Standard Deviation Non Zero', '0.4898979485566356'], ['Edges - Q1 Non Zero', '2.0'], ['Edges - Median Non Zero', '2.0'], ['Edges - Q3 Non Zero', '3.0']]
    returned = list(get_stats_from_matrix(net))
    assert expected == returned

def test_get_primary_stats():
    listt = [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3]
    listt = np.array(listt).reshape((1,len(listt)))

    matrix = np.array([[0, 0, 0, 0],[ 1, 1, 1, 1], [2, 2, 2, 2], [3, 3, 3, 3]])

    expected = {'count': 16, 'countNonZero': 12, 'sum': 24, 'max': 3, 'min': 0, 'minNonZero': 1, 'q1': 0.75, 'median': 1.5, 'q3': 2.25, 'q1NonZero': 1.0, 'medianNonZero': 2.0, 'q3NonZero': 3.0, 'average': 1.5, 'averageNonZero': 2.0, 'sumDevs': 20.0, 'sumDevsNonZero': 8.0, 'variance': 1.25, 'varianceNonZero': 0.6666666666666666, 'standardDeviation': 1.118033988749895, 'standardDeviationNonZero': 0.816496580927726}
    stats_list = get_primary_stats(listt)
    stats_matrix = get_primary_stats(matrix)
    assert expected == stats_list
    assert expected == stats_matrix

def test_get_connection_number(net):
    expected = np.array([2,3,3,2,2])
    returned = get_connection_number(net) 
    assert (expected == returned).all()

def test_get_quartiles():
    numbers = [1,2,3,4,5,6,7,8,9,10]
    expected = {'median': 5.5, 'q1': 3.25, 'q3': 7.75}
    returned = get_quartiles(numbers, None)
    assert expected == returned

def test_get_composed_stats(): #Tested in test_get_primary_stats
    assert True, "Helper function tested in test_get_primary_stats"

# Filtering
####################################

def test_binarize_mat():
    input_matrix = np.array([[3, 0, 0, 0],[0, 0, 7, 8],[0, 10, 0, 12],[0, 14, 15, 0]])
    expected_matrix = np.array([[1, 0, 0, 0],[0, 0, 1, 1],[0, 1, 0, 1],[0, 1, 1, 0]])
    returned_matrix = binarize_mat(input_matrix)
    assert (expected_matrix == returned_matrix).all()

def test_filter_cutoff_mat(matrix):
    expected_cutoff_5 = np.array([[0, 0, 0, 0],[5, 0, 7, 8],[9, 10, 0, 12],[13, 14, 15, 0]])
    expected_cutoff_14 = np.array([[0, 0, 0, 0],[0, 0, 0, 0],[0, 0, 0, 0],[0, 14, 15, 0]])
    expected_cutoff_20 = np.array([[0, 0, 0, 0],[0, 0, 0, 0],[0, 0, 0, 0],[0, 0, 0, 0]])
    
    returned_cutoff_5 = filter_cutoff_mat(matrix, 5)
    returned_cutoff_14 = filter_cutoff_mat(matrix, 14)
    returned_cutoff_20 = filter_cutoff_mat(matrix, 20)

    assert (expected_cutoff_5 == returned_cutoff_5).all()
    assert (expected_cutoff_14 == returned_cutoff_14).all()
    assert (expected_cutoff_20 == returned_cutoff_20).all()

def test_percentile_filter(matrix):
    expected_50 = np.array([[3, 0, 0, 0],[0, 0, 7, 8],[0, 10, 0, 12],[0, 14, 15, 0]])
    returned_50 = percentile_filter(matrix, 50)
    assert (expected_50 == returned_50).all()

def test_remove_zero_lines():
    input_matrix = np.array([[0, 0, 0, 0],
                             [0, 0, 0, 0],
                             [0, 0, 0, 12],
                             [0, 0, 15, 0]])
    colnames = ['S1', 'S2', 'S3', 'S4']
    rownames = ['P1', 'P2', 'P3', 'P4']
    
    expected_matrix = np.array([[0, 12], [15, 0]])
    expected_colnames = ['S3', 'S4']
    expected_rownames = ['P3', 'P4']

    returned_matrix, returned_colnames, returned_rownames = remove_zero_lines(input_matrix, colnames, rownames)
    assert (expected_matrix == returned_matrix).all()
    assert expected_colnames == returned_colnames
    assert expected_rownames == returned_rownames

# Clustering
####################################

def test_get_hc_clusters(mutation_like_branched_matrix):
    ### Starting from observation matrix (cols as variables and rows as observations)
    expected_clusters = np.array([[0],
                                  [0],
                                  [1],
                                  [2],
                                  [0]],)
    expected_clusters_objects = {'link': np.array([[0., 1., 8., 2.],
                                                   [4., 5., 8., 3.],
                                                   [3., 6., 8., 4.],
                                                   [2., 7., 8., 5.]]), 
                                  'cls': expected_clusters}
    clusters, cls_objects = get_hc_clusters(mutation_like_branched_matrix, identify_clusters = 'cut_tree', n_clusters= 3)

    assert (expected_clusters == clusters).all()
    assert (expected_clusters_objects['link'] == cls_objects['link']).all()
    assert (expected_clusters_objects['cls'] == cls_objects['cls']).all()

    ### Starting from distance matrix 
    # Using absolute distance matrix
    distance_matrix = coords2dis(mutation_like_branched_matrix)
    clusters, cls_objects = get_hc_clusters(distance_matrix, dist="custom", identify_clusters = 'cut_tree', n_clusters= 3)
    assert (expected_clusters == clusters).all()
    assert (expected_clusters_objects['link'] == cls_objects['link']).all()
    assert (expected_clusters_objects['cls'] == cls_objects['cls']).all()

    # Using normalized distance matrix
    distance_matrix = (distance_matrix - np.min(distance_matrix)) / (np.max(distance_matrix) - np.min(distance_matrix))
    expected_clusters = np.array([[8],
                                  [8],
                                  [8],
                                  [8],
                                  [8]],)
    expected_links = np.array( [[0. , 1. , 0.5, 2. ],
                                [4. , 5. , 0.5, 3. ],
                                [3. , 6. , 0.5, 4. ],
                                [2. , 7. , 0.5, 5. ]])
    clusters, cls_objects = get_hc_clusters(distance_matrix, dist="custom", identify_clusters = 'max_avg')
    assert (expected_clusters == clusters).all()
    assert (expected_links == cls_objects['link']).all()
    assert (expected_clusters == cls_objects['cls']).all()

    with pytest.raises(ValueError):
        get_hc_clusters(mutation_like_branched_matrix, identify_clusters="invalid_method")

def test_get_cls_max_avg():
    pass #This function is helper for get_hc_clusters

def test_get_avg_value4tree_cut():
    pass #This function is helper for get_hc_clusters

def test_get_member_list():
    pass #This function is helper for get_hc_clusters

def test_add_cluster_members():
    pass #This function is helper for get_hc_clusters

def test_get_cluster_avg_value():
    pass #This function is helper for get_hc_clusters

def test_tag_items_in_clusters():
    clusters = np.array([[0,0,0,0],[1,0,0,0],[2,1,1,0],[3,2,1,0]])
    names = ['C1', 'C2', 'C3', 'C4']

    expected = {0: ['C1'], 1: ['C2'], 2: ['C3'], 3: ['C4']}
    returned = tag_items_in_clusters(clusters, names)
    assert expected == returned

####################################
# Explicit conversions
####################################

# General conversor #

def test_transform2obj():
    matrix_format = np.array(  [[0, 1, 1, 0],
                                [1, 0, 0, 1],
                                [1, 0, 0, 1],
                                [0, 1, 1, 0]])
    pairs_format = [['S1', 'S2', 1], ['S1', 'S3', 1], ['S2', 'S4', 1], ['S3', 'S4', 1]]
    nested_pairs_format = {'S1': {'S2': 1, 'S3': 1}, 'S2': {'S4': 1}, 'S3': {'S4': 1}}

    returned, rowIds, colIds = transform2obj(pairs_format, inFormat="pair", outFormat="matrix")
    assert (matrix_format == returned).all()
    assert rowIds == ['S1', 'S2', 'S3', 'S4']
    assert colIds == ['S1', 'S2', 'S3', 'S4']

    returned, rowIds, colIds = transform2obj(nested_pairs_format, inFormat="nested_pairs", outFormat="matrix")
    assert (matrix_format == returned).all()
    assert rowIds == ['S1', 'S2', 'S3', 'S4']
    assert colIds == ['S1', 'S2', 'S3', 'S4']

    returned, rowIds, colIds = transform2obj(nested_pairs_format, inFormat="nested_pairs", outFormat="pair")
    assert pairs_format == returned


    expected_pairs = [['S1', 'S1', 0], ['S1', 'S2', 1], ['S1', 'S3', 1], ['S1', 'S4', 0], 
                      ['S2', 'S1', 1], ['S2', 'S2', 0], ['S2', 'S3', 0], ['S2', 'S4', 1], 
                      ['S3', 'S1', 1], ['S3', 'S2', 0], ['S3', 'S3', 0], ['S3', 'S4', 1], 
                      ['S4', 'S1', 0], ['S4', 'S2', 1], ['S4', 'S3', 1], ['S4', 'S4', 0]]
    returned, rowIds, colIds = transform2obj(matrix_format, inFormat="matrix", outFormat="pair", 
                                             colIds=['S1', 'S2', 'S3', 'S4'], rowIds=['S1', 'S2', 'S3', 'S4'])

    assert expected_pairs == returned
    assert rowIds == ['S1', 'S2', 'S3', 'S4']
    assert colIds == ['S1', 'S2', 'S3', 'S4']

    with pytest.raises(ValueError):
        transform2obj(matrix_format, inFormat="matrix", outFormat="invalid_format")
    

# Dic -> Matrix(array)
####################################

def test_to_bmatrix():
    bhash_squared = {'S1' : ['S3'], 'S2' : ['S3', 'S4'], 'S3' : ['S1'], 'S4' : ['S2', 'S3']}
    result_bmatrix_squared_to_hash = np.array([[1, 0, 0, 0], [1, 1, 0, 0], [0, 0, 1, 0], [1, 0, 0, 1]])
    test_bmatrix_squared, y_names, x_names = to_bmatrix(bhash_squared)
    assert (result_bmatrix_squared_to_hash == test_bmatrix_squared).all()
    assert y_names == ['S1', 'S2', 'S3', 'S4']
    assert x_names == ['S3', 'S4', 'S1', 'S2']

def test_to_wmatrix_squared():
    #Not symmetric case scenario
    whash_squared = {'S1' : {'S3': 3}, 'S2' : {'S3': 2, 'S4': 5}, 'S3' : {'S4': 4}} 
    result_whash_squared_to_hash = np.array([[0, 3, 0, 0], 
                                             [0, 0, 0, 4], 
                                             [0, 2, 0, 5], 
                                             [0, 0, 0, 0]])
    
    test_whash_squared, names = to_wmatrix(whash_squared, squared=True, symm=False)
    assert (result_whash_squared_to_hash == test_whash_squared).all()
    assert names == ['S1', 'S3', 'S2', 'S4']

    #Symmetric case scenario
    result_whash_squared_to_hash_symmetric = np.array([[0, 3, 0, 0], 
                                                       [3, 0, 2, 4], 
                                                       [0, 2, 0, 5], 
                                                       [0, 4, 5, 0]])
    test_whash_squared_symmetric, names = to_wmatrix(whash_squared, squared=True, symm=True)
    assert (result_whash_squared_to_hash_symmetric == test_whash_squared_symmetric).all()
    assert names == ['S1', 'S3', 'S2', 'S4']
        

def test_to_wmatrix_rectangular():
    #Not symmetric case scenario
    whash_rectangular = {'S1' : {'P3': 3}, 'S2' : {'P3': 2, 'P4': 5}, 'S3' : {'P1': 7, 'P2': 1}}
    result_whash_rectangular_to_hash = np.array([[3, 0, 0, 0], [2, 5, 0, 0], [0, 0, 7, 1]])
    test_whash_rectangular, y_names, x_names = to_wmatrix(whash_rectangular, squared=False, symm=False)
    
    assert (result_whash_rectangular_to_hash == test_whash_rectangular).all()
    assert y_names == ["S1", "S2", "S3"]
    assert x_names == ["P3", "P4", "P1", "P2"]

# List -> Dic
##################################

def test_list2dic():
    pairs = [("A","B",1), ("JJ", "J", 3.14), ("JJ", "B", 2.1412)]
    result_pair_nesteddic = list2dic(pairs)
    assert result_pair_nesteddic == {"A": {"B": 1}, "JJ":{"J": 3.14, "B": 2.1412}}

# Dic -> List
####################################

def test_nested_pairs2pairs():
    nested_pairs = {"A": {"B": 1, "C": 2}, "B": {"C": 3, "D":4}, "C":{"D": 5}}
    
    expected = [["A", "B", 1], ["A", "C", 2], ["B", "C", 3], ["B", "D", 4], ["C", "D", 5]]
    returned = nested_pairs2pairs(nested_pairs)
    assert expected == returned

# List -> Matrix(Array)
####################################

def test_pairs2matrix():
    expected_matrix_format_symmetric = np.array([[0, 1, 1, 0],
                                                 [1, 0, 0, 1],
                                                 [1, 0, 0, 1],
                                                 [0, 1, 1, 0]])
    expected_matrix_format_asymmetric = np.array([[1. ,1. ,0.],
                                                  [0. ,0. ,1.],
                                                  [0. ,0. ,1.]])
    
    pairs_format = [['S1', 'S2', 1], ['S1', 'S3', 1], ['S2', 'S4', 1], ['S3', 'S4', 1]]
    returned_matrix, rowIds, colIds = pairs2matrix(pairs_format)

    assert ( returned_matrix == expected_matrix_format_symmetric).all()
    assert rowIds == ['S1', 'S2', 'S3', 'S4']
    assert colIds == ['S1', 'S2', 'S3', 'S4']

    asym_returned_matrix, rowIds, colIds = pairs2matrix(pairs_format, symm=False)
    assert (asym_returned_matrix == expected_matrix_format_asymmetric).all()
    assert rowIds == ['S1', 'S2', 'S3']
    assert colIds == ['S2', 'S3', 'S4']

def test_update_index(): #TODO: not quite sure if this is how the code is actually working, but this function is also being tested in higher order functions
    index = {"A": 2, "B":3}
    assert update_index(index, "A", 2) == 2
    assert update_index(index, "B", 3) == 3
    assert update_index(index, "C", 0) == 1


# Matrix(Array) -> List
####################################

def test_matrix2pairs(net, rectangular_matrix, rowIds, colIds):
    expected_rectangular_pairs = [['P1', 'D1', 1], ['P1', 'D2', 1], ['P1', 'D3', 0], 
                                  ['P2', 'D1', 0], ['P2', 'D2', 0], ['P2', 'D3', 1], 
                                  ['P3', 'D1', 0], ['P3', 'D2', 0], ['P3', 'D3', 1], 
                                  ['P4', 'D1', 1], ['P4', 'D2', 1], ['P4', 'D3', 0], 
                                  ['P5', 'D1', 1], ['P5', 'D2', 1], ['P5', 'D3', 1]]
    returned_rectangular_pairs = matrix2pairs(rectangular_matrix, rowIds, colIds)
    assert expected_rectangular_pairs == returned_rectangular_pairs

    expected_symm_pairs = [['P1', 'P1', 0], ['P1', 'P2', 1], ['P1', 'P3', 1], ['P1', 'P4', 0], ['P1', 'P5', 0], 
                                            ['P2', 'P2', 0], ['P2', 'P3', 1], ['P2', 'P4', 0], ['P2', 'P5', 1], 
                                                             ['P3', 'P3', 0], ['P3', 'P4', 1], ['P3', 'P5', 0], 
                                                                              ['P4', 'P4', 0], ['P4', 'P5', 1], 
                                                                                               ['P5', 'P5', 0]]
    returned_squared_symmetric_pairs = matrix2pairs(net, rowIds, rowIds, symm=True)
    assert expected_symm_pairs == returned_squared_symmetric_pairs

def test_matrixes2pairs(net, rectangular_matrix, rowIds, colIds):
    doubled_rectangular = rectangular_matrix * 2
    doubled_symmetric = net * 2

    expected_rectangular_pairs = [['P1', 'D1', 1, 2], ['P1', 'D2', 1, 2], ['P1', 'D3', 0, 0], 
                                  ['P2', 'D1', 0, 0], ['P2', 'D2', 0, 0], ['P2', 'D3', 1, 2], 
                                  ['P3', 'D1', 0, 0], ['P3', 'D2', 0, 0], ['P3', 'D3', 1, 2], 
                                  ['P4', 'D1', 1, 2], ['P4', 'D2', 1, 2], ['P4', 'D3', 0, 0], 
                                  ['P5', 'D1', 1, 2], ['P5', 'D2', 1, 2], ['P5', 'D3', 1, 2]]
    returned_rectangular_pairs = matrixes2pairs([rectangular_matrix, doubled_rectangular], rowIds, colIds)
    assert expected_rectangular_pairs == returned_rectangular_pairs


    expected_symm_pairs = [['P1', 'P1', 0, 0], ['P1', 'P2', 1, 2], ['P1', 'P3', 1, 2], ['P1', 'P4', 0, 0], ['P1', 'P5', 0, 0], 
                                               ['P2', 'P2', 0, 0], ['P2', 'P3', 1, 2], ['P2', 'P4', 0, 0], ['P2', 'P5', 1, 2], 
                                                                   ['P3', 'P3', 0, 0], ['P3', 'P4', 1, 2], ['P3', 'P5', 0, 0], 
                                                                                       ['P4', 'P4', 0, 0], ['P4', 'P5', 1, 2], 
                                                                                                           ['P5', 'P5', 0, 0]]
    returned_symmetric_pairs = matrixes2pairs([net, doubled_symmetric], rowIds, rowIds, symm=True)
    assert expected_symm_pairs == returned_symmetric_pairs
    

def test_matrix2relations():
    rectangular_matrix = np.array([[-0.1, -3.0,  2.0],
                                   [ 1.0, -4.0, -1.0]])
    rowIds = ['P1', 'P2']
    colIds = ['D1', 'D2', 'D3']
    expected_relations = [["P1", "D3", 2], ["P2", "D1", 1]]
    
    returned_rectangular_pairs = matrix2relations(rectangular_matrix, rowIds, colIds)
    assert expected_relations == returned_rectangular_pairs

    symm_matrix = np.array([[-1.0,  1.0, -1.0],
                            [ 1.0, -1.0,  1.0],
                            [-1.0,  1.0, -1.0]])

    expected_symm_pairs = [["D1", "D2", 1], ["D2", "D3", 1]]
    returned_squared_symmetric_pairs = matrix2relations(symm_matrix, colIds, colIds, symm=True)
    assert expected_symm_pairs == returned_squared_symmetric_pairs


import numpy as np
import pandas as pd
from sklearn.metrics import (
    roc_auc_score, 
    average_precision_score
)

class Performancer:

    def __init__(self):
        self.control = {}       # id -> y_true
        self.predictions = []   # [(id, score)]
        self.y_true = []
        self.y_score = []
        self.factors = [] # e.g. "algorithm1", "configset2",...
        self.metric_table = []

    def load_controls(self, ref_array):
        for _id, y in ref_array:
            self.control[_id] = float(y)

    def load_predictions(self, pred_array):
        self.predictions = pred_array

    def _align_y(self):
        y_true, y_score = [], []
        for _id, s in self.predictions:
            if _id in self.control:
                y_true.append(int(self.control[_id]))
                y_score.append(float(s))
        self.y_true = np.asarray(y_true, dtype=int)
        self.y_score = np.asarray(y_score, dtype=float)

    def metrics_table_by_score( self, metrics=("tpr", "fnr", "tnr", "fpr"), descending=True, 
                               score_cut=np.inf, add_factors=False, add_counts=False):
        if len(self.y_true) == 0: self._align_y()

        y_true = np.asarray(self.y_true, dtype=int)
        y_score = np.asarray(self.y_score, dtype=float)

        n = len(y_true)

        order = np.argsort(-y_score) if descending else np.argsort(y_score)
        s = y_score[order]
        y = y_true[order]

        P = int((y == 1).sum())
        N = n - P
        tp_cum = np.cumsum(y == 1)
        fp_cum = np.cumsum(y == 0)

        change_idx = np.r_[np.where(np.diff(s) != 0)[0], n - 1]
        thr = s[change_idx]
        tp = tp_cum[change_idx].astype(float)
        fp = fp_cum[change_idx].astype(float)
        fn = float(P) - tp
        tn = float(N) - fp

        def safe_div(num, den):
            return np.divide(num, den, out=np.zeros_like(num, dtype=float), where=den > 0)

        P_f = float(P)
        N_f = float(N)
        T_f = float(n)

        # Primary metrics
        tpr = safe_div(tp, P_f)
        fnr = safe_div(fn, P_f)
        tnr = safe_div(tn, N_f)
        fpr = safe_div(fp, N_f)
        # Secondary metrics
        prec = safe_div(tp, tp + fp)
        f1 = safe_div(2 * tp, 2 * tp + fp + fn)
        acc = safe_div(tp + tn, T_f)
        bal_acc = 0.5 * (tpr + tnr)
        denom = np.sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn))
        mcc = safe_div(tp * tn - fp * fn, denom)
        coverage = safe_div(tp + fp, T_f)
        cols = {
            "score": thr.astype(float),
            "tpr": tpr, "fnr": fnr, "tnr": tnr, "fpr": fpr,
            "prec": prec, "f1": f1,
            "acc": acc, "bal_acc": bal_acc,
            "mcc": mcc,
            "coverage": coverage,
        }
        if add_counts:
            cols.update({"tp": tp, "fp": fp, "fn": fn, "tn": tn})

        df = pd.DataFrame(cols)

        keep = ["score"]
        for m in metrics:
            if m not in df.columns:
                raise KeyError(f"Metric '{m}' not available. Availables: {list(df.columns)}")
            keep.append(m)
        if add_counts:
            keep += ["tp", "fp", "fn", "tn"]

        df = df[keep]

        if score_cut is None or score_cut is np.inf:
            pass 
        else:
            k = int(score_cut)
            if k > 0 and k < len(df):
                idx = np.linspace(0, len(df) - 1, num=k, dtype=int)
                idx = np.unique(idx)
                df = df.iloc[idx].reset_index(drop=True)

        df = df.sort_values("score", ascending=not descending, ignore_index=True)
        if add_factors and self.factors:
            for i, f in enumerate(self.factors, start=1):
                df[f"factor{i}"] = f

        self.metric_table = df

    def get_auc_metrics(self, descending=True):
        if len(self.y_true) == 0: self._align_y()
        if len(np.unique(self.y_true)) != 2:
            return {"roc_auc": np.nan, "pr_auc": np.nan}
        c = 1 if descending else -1
        
        return {
            "roc_auc": float(roc_auc_score(self.y_true, c * self.y_score)),
            "pr_auc": float(average_precision_score(self.y_true, c * self.y_score)),
        }

    def get_best_thresholds_from_table(self, best_for=("f1", "mcc")):
        out = {}
        for m in best_for:
            col = self.metric_table[m].to_numpy(dtype=float)
            idx = int(np.nanargmax(col))
            out[f"{m}_best"] = float(self.metric_table.iloc[idx][m])
            out[f"{m}_best_threshold"] = float(self.metric_table.iloc[idx]["score"])
        return out

    def get_summary_metrics(self, best_for=("f1", "mcc"), descending=True, add_factors=False):
        out = {}
        out.update(self.get_auc_metrics(descending))
        out.update(self.get_best_thresholds_from_table(best_for=best_for))
        if add_factors and self.factors:
            for i, f in enumerate(self.factors, start=1):
                out[f"factor{i}"] = f
        return out


import argparse, sys

from py_cmdtabs.cmdtabs import CmdTabs
from py_exp_calc.main_modules import *

def based_0(string): return int(string) - 1
def list_based_0(string): return CmdTabs.parse_column_indices(",", string)
def double_split(string, sep1=";", sep2=","):
    return [sublst.split(sep2) for sublst in string.strip().split(sep1)]

## Common options
def add_common_options(parser):
    parser.add_argument("-i", "--input_file", dest="input",
        help="Path to input file")
    parser.add_argument("-o", "--output_file", dest="output",
        help="Path to input file")

def clusterize(args = None):
    if args == None: args = sys.argv[1:]
    parser = argparse.ArgumentParser(description="Get clusters")
    add_common_options(parser)
    parser.add_argument("-x", "--x_dim", dest="x_dim",
        help="Item list file for x axis")
    parser.add_argument("-y", "--y_dim", dest="y_dim",
        help="Item list file for y axis")
    parser.add_argument("-n", "--n_clusters", dest="n_clusters", type=int,
        help="N clusters for cut tree algorithm")
    parser.add_argument("-H", "--height", dest="height",
        help="Cut the tree with cut tree algorithm at specifc height")
    parser.add_argument("-c", "--clustering", dest="clustering", default='cut_tree',
        help="Method to identify clusters in hierarchical tree")        
    parser.add_argument("-r","--report", dest="report", default= False, action="store_true",
        help="Make report. Default false")

    opts =  vars(parser.parse_args(args))
    main_clusterize(opts)


def inference_analyzer(args = None):
    if args == None: args = sys.argv[1:]
    parser = argparse.ArgumentParser(description="Perform statistics form table")
    add_common_options(parser)
    parser.add_argument("-f", "--factor_indexes", dest="factor_indexes", type= list_based_0,
        help="Specify col number for factors")
    parser.add_argument("--alt_hyp", dest="alt_hyp", default= "two-sided",
        help="Select alternative hypotesis with optinos: two-sided, less, greater")
    parser.add_argument("-np","--no_parametric", dest="no_parametric", default= False, action='store_true', 
        help="Use this option if you want to specify not parametric analysis for all tests")
    parser.add_argument("--adj_pval", dest="adj_pval", default= None,
        help="Select multiple testing adjustment: bonferroni, sidak, holm-sidak, holm, simes-hochberg, hommel, fdr_bh, fdr_by, fdr_tsbh, fdr_tsbky")
    parser.add_argument("--header", dest = "header", default=False, action='store_true',
                        help="This is to check if you got a header")
    opts =  vars(parser.parse_args(args))
    main_inference_analyzer(opts)

def stable_select(args=None):
    if args is None:
        args = sys.argv[1:]
    parser = argparse.ArgumentParser(
        description="Stable feature selection (e.g., elastic_net) from data table + label file"
    )
    add_common_options(parser)
    parser.add_argument("--label", dest="label", required=True,
        help="Path to label file (single column, no header)")
    parser.add_argument("--sep", dest="sep", default="\t",
        help="Separator for input data table. Default: TAB")
    parser.add_argument("--header", dest="header", type=int, default=0,
        help="Header row index for data table. Default: 0")
    parser.add_argument("--index_col", dest="index_col", type=int, default=0,
        help="Index column for data table. Default: 0")
    parser.add_argument("--regularization_strength", dest="regularization_strength", type=float, default=0.5,
        help="Regularization strength parameter. Default: 0.25. lowest value is 0, which implies no regularization.")
    parser.add_argument("--l1_ratio", dest="l1_ratio", type=float, default=0.01,
        help="Elastic Net mixing parameter (0=L2, 1=L1). Default: 0.05.")
    parser.add_argument("--number_iter", dest="number_iter", type=int, default=50,
        help="Number of iterations. Default: 10")
    parser.add_argument("--sample_fraction", dest="sample_fraction", type=float, default=0.8,
        help="Sample fraction per iteration. Default: 0.8")
    parser.add_argument("--proportion_threshold", dest="proportion_threshold", type=float, default=0.5,
        help="Stability proportion threshold. Default: 0.5")
    parser.add_argument("--random_state", dest="random_state", type=int, default=0,
        help="Random state. Default: 0")
    parser.add_argument("--return_stability", dest="return_stability",
        default=False, action="store_true",
        help="If set, return stability metrics (depends on stable_selection implementation). Default: False")

    opts = vars(parser.parse_args(args))
    main_stable_select(opts)


def performancer(args = None):
    if args == None: args = sys.argv[1:]
    parser = argparse.ArgumentParser(description="Get clusters")
    add_common_options(parser)
    parser.add_argument("-c","--control_file", dest="control_file", default=None,
        help="Control file name")
    parser.add_argument("--metrics", dest="metrics", default=["tpr", "fnr", "tnr", "fpr", "coverage", "prec", "f1", "mcc"], 
                        type = lambda x: x.strip().split(","),
                        help= """Add the different metrics necesssary to 
                        evaluate the system: acc, f1, prec, tpr, fnr, tnr, fpr, bal_acc, mcc""")
    parser.add_argument("--generate_report", dest= "generate_report", default= False, action="store_true")
    parser.add_argument("--generate_table", dest = "generate_table", default = False, action="store_true")
    parser.add_argument("--ascending_order", dest= "ascending", default=False, action="store_true" )
    opts =  vars(parser.parse_args(args))
    main_performancer(opts)


import os, time
start = time.time()
from importlib.resources import files
from scipy.spatial.distance import squareform

from py_report_html.py_report_html import Py_report_html
from py_cmdtabs.cmdtabs import CmdTabs
from py_exp_calc.exp_calc import *

def main_stable_select(args):
    import pandas as pd
    data = pd.read_csv(
        args["input"],
        sep=args["sep"],
        header=args["header"],
        index_col=args["index_col"]
    )
    label = pd.read_csv(args["label"], header=None)

    new_table = stable_selection(
        data,
        label,
        regularization_strength=args["regularization_strength"],
        l1_ratio=args["l1_ratio"],
        number_iter=args["number_iter"],
        sample_fraction=args["sample_fraction"],
        proportion_threshold=args["proportion_threshold"],
        random_state=args["random_state"],
        return_stability=args["return_stability"]
    )

    new_table.to_csv(args["output"], sep="\t", index=True, header=True)

def main_inference_analyzer(args):
    table = CmdTabs.load_input_data(args['input'])
    table = parse_table(table, header=args['header'], factor_index= args['factor_indexes'])
    stats = get_test(table, args['factor_indexes'], alternative = args['alt_hyp'], header = args["header"], adj_pval=args['adj_pval'], parametric = not args['no_parametric'])
    CmdTabs.write_output_data(stats, output_path=args['output'])

def main_clusterize(args):
    import numpy as np

    observation_matrix, y_names, x_names = load(args['input'], 
        x_axis_file=args['x_dim'], y_axis_file=args['y_dim'])


    if len(observation_matrix.shape) == 1: # transform 1 Dimensional array to 2 Dimensional array
        observation_matrix = observation_matrix.reshape(observation_matrix.size, 1)
    if args['clustering'] == 'kmeans':
        from sklearn.cluster import KMeans
        kmeans = KMeans(n_clusters=args['n_clusters'], random_state=0, n_init="auto").fit(observation_matrix)
        with open(args['output'], 'w') as f:
            for i, cluster_id in enumerate(kmeans.labels_): f.write(f"{x_names[i]}\t{cluster_id}\n")
    elif args['clustering'] == 'mbk':
        from sklearn.cluster import MiniBatchKMeans
        mbk = MiniBatchKMeans(init='k-means++', n_clusters=args['n_clusters'], batch_size=1024,
                      n_init=10, max_no_improvement=10, verbose=0).fit(observation_matrix)
        with open(args['output'], 'w') as f:
            for i, cluster_id in enumerate(mbk.labels_): f.write(f"{x_names[i]}\t{cluster_id}\n")
    else:
        clusters, cls_objects = get_hc_clusters(observation_matrix, identify_clusters=args['clustering'], item_list = x_names, n_clusters= args['n_clusters'])
        write_dict(clusters, args['output'])
        if args['report']:
            template = open(str(files('py_exp_calc.templates').joinpath('clustering.txt'))).read()
            container = {   'dist': squareform(cls_objects['dist_vector']), 
                            'link': cls_objects['link'], 
                            'raw_cls': cls_objects['cls'] }
            report = Py_report_html(container, 'Clustering')
            report.build(template)
            report.write(os.path.join(os.path.dirname(args["output"]), 'clustering.html'))

def read_tabular_file(input_file, cols):
    data = []
    with open(input_file) as f:
        for line in f:
            fields = line.rstrip().split("\t")
            data.append([fields[i] for i in cols])
    return data

def write_dict(dict, file):
    with open(file, 'w') as f:
        for k, values in dict.items():
            f.write(f"{k}\t{','.join(values)}\n")

def parse_table(table, header, factor_index):
    parsed_table = []
    if header:
        header = table.pop(0)
    for row in table:
        parsed_row = []
        for idx, el in enumerate(row):
            if not idx in factor_index:
                parsed_row.append(float(el))
            else:
                parsed_row.append(el)
        parsed_table.append(parsed_row)
    table = parsed_table
    if header:
        table.insert(0, header)
    return table


def main_performancer(options):
    from py_exp_calc.performancer import Performancer
    controls = get_records(options['control_file'])
    predictions = get_records(options['input'])
    sample_ids = sorted(set([row[2] for row in controls]))
    metric_table = []
    summary_table = []
    for sample_id in sample_ids:
        controls_subset = [[row[0], row[1]] for row in controls if row[2] == sample_id]
        predictions_subset = [[row[0], row[1]] for row in predictions if row[2] == sample_id]
        factors = []
        if len(predictions[0]) > 3:
            factors = [row[3:] for row in predictions if row[2] == sample_id]
            factors = factors[0]
        performancer = Performancer()
        performancer.load_controls(controls_subset)
        performancer.load_predictions(predictions_subset)
        performancer.factors = factors
        # Creating metrics by score
        performancer.metrics_table_by_score( metrics=options["metrics"], descending=not options["ascending"], score_cut=np.inf, 
                                            add_factors=True, add_counts=False)
        df = performancer.metric_table
        df.insert(0, "sample_id", sample_id)
        if not metric_table:
            metric_table = [df.columns.tolist()]
        metric_table.extend(df.values.tolist())

        # Creating summary score
        summary = performancer.get_summary_metrics(
            best_for=options["metrics"], descending=not options["ascending"],
            add_factors=True
        )
        summary["sample_id"] = sample_id
        if not summary_table:
            factor_cols = sorted(k for k in summary.keys() if k.startswith("factor"))
            metric_cols = sorted(k for k in summary.keys() 
                                 if k not in factor_cols and k != "sample_id")
            header = ["sample_id"] + metric_cols + factor_cols
            summary_table.append(header)
        row_summary = [summary.get(col) for col in summary_table[0]]
        summary_table.append(row_summary)

    if options["generate_table"]:
        with open(options["output"] + "_table", "w") as f:
            for row in metric_table:
                row = [str(col) for col in row]
                f.write("\t".join(row) + "\n")

        with open(options["output"] + "_summary_table", "w") as f:
            for row in summary_table:
                row = [str(col) for col in row]
                f.write("\t".join(row) + "\n")

    if options["generate_report"]:
        template = open(str(files('py_exp_calc.templates').joinpath('performancer.txt'))).read()
        container = {'metrics_table': metric_table, 'summary_table': summary_table, 'metrics': options['metrics']}
        report = Py_report_html(container, 'Performance')
        report.build(template)
        report.write(options["output"]+'_report.html')

def get_records(file):
    return CmdTabs.load_input_data(file, autodetect_compression=True)



"""
This is a skeleton file that can serve as a starting point for a Python
console script. To run this script uncomment the following lines in the
``[options.entry_points]`` section in ``setup.cfg``::

    console_scripts =
         fibonacci = py_exp_calc.skeleton:run

Then run ``pip install .`` (or ``pip install -e .`` for editable mode)
which will install the command ``fibonacci`` inside your current environment.

Besides console scripts, the header (i.e. until ``_logger``...) of this file can
also be used as template for Python modules.

Note:
    This file can be renamed depending on your needs or safely removed if not needed.

References:
    - https://setuptools.pypa.io/en/latest/userguide/entry_point.html
    - https://pip.pypa.io/en/stable/reference/pip_install
"""
import sys, copy, os, itertools
from io import BytesIO
import numpy as np
from scipy.spatial import distance_matrix
from scipy.cluster.hierarchy import linkage, to_tree, cut_tree
from scipy import sparse
from scipy.sparse import save_npz, load_npz
from scipy import stats


####################################
# LIST OPERATIONS
####################################

# tree formats
####################################

def transform_tree(tree, source, target, **options):
  if source == "r" and target == "python":
    target_tree =  r2python_tree(tree)
  elif source == "python" and target == "newick":
    scipy_tree_obj = to_tree(tree, False)
    target_tree = get_newick(scipy_tree_obj, scipy_tree_obj.dist, options.get('leaf_names'))
  else:
    raise Exception(f"Conversion method from {source} to {target} not implemented yet")
  return target_tree

def get_newick(node, parent_dist, leaf_names=[], newick='') -> str:
  """
  Convert sciply.cluster.hierarchy.to_tree()-output to Newick format.

  :param node: output of sciply.cluster.hierarchy.to_tree()
  :param parent_dist: output of sciply.cluster.hierarchy.to_tree().dist
  :param leaf_names: list of leaf names
  :param newick: leave empty, this variable is used in recursion.
  :returns: tree in Newick format
  """
  if node.is_leaf():
      nodeID = leaf_names[node.id] if leaf_names else node.id
      return "%s:%.2f%s" % (nodeID, parent_dist - node.dist, newick)
  else:
      if len(newick) > 0:
          newick = "):%.2f%s" % (parent_dist - node.dist, newick)
      else:
          newick = ");"
      newick = get_newick(node.get_left(), node.dist, leaf_names, newick=newick)
      newick = get_newick(node.get_right(), node.dist, leaf_names, newick=",%s" % (newick))
      newick = "(%s" % (newick)
      return newick

def r2python_tree(rtree):
  n=0
  for ida, idb, _ in rtree:
    if -ida > n:
      n=-ida 
    if -idb > n:
      n=-idb
  ptree=[]
  for ida, idb, dis in rtree:
    row = [ida,idb,dis]
    for idx in [0,1]:
      if row[idx] < 0:
        row[idx] = -row[idx] - 1 # R starts in 1
      else:
        row[idx] = row[idx] + n - 1
    ptree.append(row)
  ptree = add_elements_by_cluster(ptree,n)
  return ptree

def add_elements_by_cluster(ptree,n):
  complete_ptree = []
  for i,iteration in enumerate(ptree):
    els=0
    for idx in [0,1]:
      index_val = iteration[idx]
      if index_val < n:
        els+=1
      else:
        els+=complete_ptree[int(index_val-n)][3]
    complete_ptree.append([iteration[0],iteration[1],iteration[2], els])
  return complete_ptree

# One List operation
####################################

def get_stats_from_list(data): #TODO: think inj a dry version #DONE
  data = np.array(data).reshape((1,len(data)))
  primary_stats = get_primary_stats(data)
  stats = []
  stats.append(['Elements', primary_stats["count"]])
  stats.append(['Elements Non Zero', primary_stats["countNonZero"]])
  stats.append(['Non Zero Density', primary_stats["countNonZero"]/primary_stats["count"]])
  stats.append(['Max', primary_stats["max"]])
  stats.append(['Min', primary_stats["min"]])
  stats.append(['Average', primary_stats["average"]])
  stats.append(['Variance', primary_stats["variance"]])
  stats.append(['Standard Deviation', primary_stats["standardDeviation"]])
  stats.append(['Q1', primary_stats["q1"]])
  stats.append(['Median', primary_stats["median"]])
  stats.append(['Q3', primary_stats["q3"]])
  stats.append(['Min Non Zero', primary_stats["minNonZero"]])
  stats.append(['Average Non Zero', primary_stats["averageNonZero"]])
  stats.append(['Variance Non Zero', primary_stats["varianceNonZero"]])
  stats.append(['Standard Deviation Non Zero', primary_stats["standardDeviationNonZero"]])
  stats.append(['Q1 Non Zero', primary_stats["q1NonZero"]])
  stats.append(['Median Non Zero', primary_stats["medianNonZero"]])
  stats.append(['Q3 Non Zero', primary_stats["q3NonZero"]])
  stats = map(lambda x: [x[0],str(x[1])],stats)
  return stats

def uniq(arr1):
  used = set()      
  return [x for x in arr1 if x not in used and (used.add(x) or True)]

def flatten(arr):
    result = []
    if type(arr) is not list:
        return [arr]
    elif len(arr) == 0: 
        return []
    elif len(arr) == 1: 
        return flatten(arr[0]) 
    elif len(arr) > 1:
        for sub_arr in arr: 
            result += flatten(sub_arr)
        return result

def add_tags(table, tags, keys, default_tag = None):
  tagged_file=[]
  for row in table: 
    record = dig(tags, *tuple([row[key] for key in keys]))
    if record is not None:
      row.append(record)
      tagged_file.append(row)
    elif default_tag:
      row.append(default_tag)
      tagged_file.append(row)
  return tagged_file

# Variable seleccition
####################################

def stable_selection(data, label, l1_ratio=0.5, regularization_strength=1.0,
                     number_iter=300, sample_fraction=0.8,
                     proportion_threshold=0.7, random_state=0,
                     return_stability=False):
    stability = get_stability(
        data, label,
        l1_ratio=l1_ratio, regularization_strength=regularization_strength,
        number_iter=number_iter, sample_fraction=sample_fraction,
        random_state=random_state
    )

    # DataFrame
    if hasattr(data, "loc") and hasattr(data, "columns"):
        keep = stability.index[stability >= proportion_threshold]
        data_mod = data.loc[:, keep]
        return (data_mod, keep.tolist(), stability) if return_stability else data_mod

    # numpy array
    keep = np.where(stability >= proportion_threshold)[0]
    data_mod = np.asarray(data)[:, keep]
    return (data_mod, keep.tolist(), stability) if return_stability else data_mod

  
def get_stability(data, label, l1_ratio=0.5, regularization_strength=1.0,
                  number_iter=300, sample_fraction=0.8, random_state=0):
    from sklearn.linear_model import LogisticRegression, ElasticNet
    from sklearn.preprocessing import StandardScaler

    try:
        names = np.array(data.columns)
        X = data.to_numpy()
        is_df = True
    except Exception:
        X = np.asarray(data)
        names = np.arange(X.shape[1])
        is_df = False

    y = np.asarray(label)
    y = label.to_numpy().ravel()
    y = (y == np.unique(y)[-1]).astype(int)  # map a 0/1

    n, p = X.shape
    m = max(2, int(np.ceil(sample_fraction * n)))
    rng = np.random.default_rng(random_state)

    model = LogisticRegression(penalty="elasticnet", solver="saga", l1_ratio=l1_ratio, C=1.0/regularization_strength, max_iter=5000)

    counts = np.zeros(p, dtype=int)

    for _ in range(number_iter):
        idx = rng.choice(n, size=m, replace=False)
        Xb = StandardScaler().fit_transform(X[idx])
        yb = y[idx]

        if np.unique(yb).size < 2:
            continue

        model.fit(Xb, yb)
        counts[np.abs(model.coef_.ravel()) > 0] += 1

    stability = counts / float(number_iter)

    if is_df:
        import pandas as pd
        return pd.Series(stability, index=names, name="stability").sort_values(ascending=False)
    return stability

# Dimensionality reduction
####################################

def data2umap(data, n_neighbors=15, min_dist=0.1, n_components=2, metric='euclidean', random_seed=None):
    import umap #Importing UMAP here because of the long loading time of the library.
    if metric == 'precomputed':
        if data.shape[0] != data.shape[1]:
            raise ValueError("Precomputed distance matrix must be square.")

    reducer = umap.UMAP(
        n_neighbors=n_neighbors,
        min_dist=min_dist,
        n_components=n_components,
        metric=metric,
        random_state=random_seed
    )
    umap_coords = reducer.fit_transform(data)
    return umap_coords

# Stats
####################################

def get_test(table, idx_factor, alternative = "two-sided", adj_pval=None, assumptions={"continuos": True, "paired": False}, header = True, binarize_factor = True, parametric = True):
    import statsmodels.api as sm # VEry slow loading
    idx_variable = diff(list(range(0,len(table[0]))), idx_factor)
    header = get_header(table, idx_factor, header, idx_variable)
    stats = []
    for factor in idx_factor:
      for variable in idx_variable:
        table_pair = [[row[index] for index in [factor,variable]] for row in table]
        factor_values = [row[factor] for row in table]
        factor_values = sorted(list(set(factor_values)))
        if len(factor_values) > 2:
          for factorA, factorB in itertools.combinations(factor_values,2):
            filtered_table_pair = [row for row in table if row[factor] in [factorA, factorB]]
            pvalue = one_hypothesis(filtered_table_pair, alternative, assumptions, parametric)
            stats.append([f"{header[factor]}_{factorA}:{factorB}", header[variable], pvalue])
        else:
          pvalue = one_hypothesis(table_pair, alternative, assumptions, parametric)
          stats.append([header[factor], header[variable], pvalue])
    if adj_pval: 
      pvalues =  [row[2] for row in stats]
      adj_pvals = sm.stats.multipletests(pvalues, method=adj_pval, is_sorted=False, returnsorted=False)[1]
      for i in range(len(pvalues)):
        stats[i].append(adj_pvals[i])
    return stats

def get_header(table, idx_factor, header, idx_variable):
    if header:
      header = table.pop(0)
    else:
      header = []
      factor_number = 1
      variable_number = 1
      for i in range(0,len(table[0])):
        if i in idx_factor:
          header.append(f"F{factor_number}")
          factor_number += 1
        if i in idx_variable:
          header.append(f"V{variable_number}")
          variable_number += 1
    return header

def one_hypothesis(table_pair, alternative, assumptions={"continuos": True, "paired": False}, parametric = True):
  # 1) Continuos or not? 2) Paired? 3) Parametric?
  if assumptions["continuos"]:
    factor_values = sorted(list(set([row[0] for row in table_pair])))
    a = [row[1] for row in table_pair if row[0] == factor_values[0]]
    b = [row[1] for row in table_pair if row[0] == factor_values[1]]
    
    if assumptions["paired"]:  
      pvalue = stats.ttest_rel(a, b, axis=0, nan_policy='propagate', alternative=alternative).pvalue
    else:
      if parametric:
        pvalue = stats.ttest_ind(a, b, axis=0, equal_var=False, 
        nan_policy='propagate',
        alternative=alternative).pvalue
      else:
        pvalue = stats.mannwhitneyu(a, b, alternative=alternative, nan_policy='propagate').pvalue
  else:
    raise Exception("Not implemented yet for categorical variables")
  return pvalue

# Combine p-values
####################################

def pvalue_bayesian_combination(p_values):
    p_values = np.clip(p_values, 1e-10, 1 - 1e-10)  
    return 1 - np.prod(1 - p_values)

def pvalue_fisher_combination(p_values):
    p_values = np.clip(p_values, 1e-10, 1)  
    chi2_stat = -2 * np.sum(np.log(p_values))
    df = 2 * len(p_values) 
    return 1 - stats.chi2.cdf(chi2_stat, df)

def pvalue_stouffer_combination(p_values, weights=None):
    if weights is None: weights = np.ones(len(p_values)) 
    p_values = np.clip(p_values, 1e-10, 1) 
    z_scores = stats.norm.ppf(1 - p_values)
    z_combined = np.sum(weights * z_scores) / np.sum(weights) 
    return 1 - stats.norm.cdf(z_combined)

# Scores
####################################

def get_rank_metrics(scores, ids, pos_selection="worst"):
    ordered_scores = []
    ordered_indexes = np.argsort(scores)  # from smallest to largest
    number_of_all_nodes = len(ids)

    last_val = None
    n_elements = ordered_indexes.shape[0]
    if pos_selection == "best":
      positions = reversed(range(0,n_elements))
    elif pos_selection == "worst":
      positions = range(0,n_elements)

    for pos in positions:
        order_index = ordered_indexes[pos]
        val = scores[order_index]
        node_name = ids[order_index]

        if pos_selection == "worst":
          rank = get_position_for_items_with_same_score(
                  pos, val, last_val, n_elements, ordered_scores)  # number of items behind
          rank = n_elements - rank  # number of nodes below or equal
        elif pos_selection == "best":
          rank = get_position_members_up(
                  pos, val, last_val, n_elements, ordered_scores)  # number of items behind or equal
        rank_percentage = rank/number_of_all_nodes

        ordered_scores.append(
                [node_name, val, rank_percentage, rank])
        last_val = val

    if pos_selection == "worst": ordered_scores.reverse()  # from largest to smallest
    ordered_scores = add_absolute_rank_column(
            ordered_scores)
        
    return ordered_scores

def get_position_members_up(pos, val, prev_val, n_elements, ordered_scores):
    members_behind = 1
    if prev_val is not None:
        if prev_val > val:
            members_behind = n_elements - pos 
        else:
            members_behind = ordered_scores[-1][3] 
    return members_behind

def get_position_for_items_with_same_score(pos, val, prev_val, n_elements, ordered_scores):
    members_behind = 0
    if prev_val is not None:
        if prev_val < val:
            members_behind = pos
        else:
            members_behind = n_elements - ordered_scores[-1][3]
    return members_behind

def add_absolute_rank_column(ranking):
    ranking_with_new_column = []
    absolute_rank = 1
    n_rows = len(ranking)
    for row_pos in range(n_rows):
        if row_pos == 0:
            new_row = ranking[row_pos] + [absolute_rank]
            ranking_with_new_column.append(new_row)
        else:
            prev_val = ranking[row_pos-1][2]
            val = ranking[row_pos][2]
            if val > prev_val:
                absolute_rank += 1

            new_row = ranking[row_pos] + [absolute_rank]
            ranking_with_new_column.append(new_row)
    return ranking_with_new_column

# List x List operation
####################################

def intersection(arr1, arr2, indexing = False):
  if indexing: 
    index_arr2 = flatlist2dic(arr2)
    intersection = [item for item in arr1 if index_arr2.get(item)] 
  else:
    intersection = [item for item in arr1 if item in arr2] 
  return intersection

def union(arr1, arr2, indexing = False):
  if indexing:
    index_arr1 = flatlist2dic(arr1)
    union = arr1 + [item for item in arr2 if not index_arr1.get(item)] 
  else:
    union = arr1 + [item for item in arr2 if item not in arr1]
  return union

def diff(arr1, arr2, indexing = False):
  if indexing:
    index_arr2 = flatlist2dic(arr2)
    diff = [item for item in arr1 if not index_arr2.get(item)]
  else:
    diff = [item for item in arr1 if item not in arr2]
  return diff

####################################
# DIC OPERATIONS
####################################

# Values - operation
####################################

## Get
######

def dig(dictio, *keys):
  try:
    for key in keys:
        dictio = dictio[key]
    return dictio
  except KeyError:
    return None 

def get_hash_values_idx(dictio):
  x_names_indx = {}
  i = 0
  for k, values in dictio.items():
    for val_id in values:
      if type(val_id) is list: val_id = val_id[0]
      query = x_names_indx.get(val_id)
      if query == None:
        x_names_indx[val_id] = i
        i += 1
  return x_names_indx

# Insert
########

# This method absorbed the add_record version of netanalyzer
def add_record(dictio, key, record, uniq=False):
  query = dictio.get(key)
  if query == None:
    dictio[key] = [record]
  elif not uniq: # We not take care by repeated entries
    query.append(record)
  elif not record in query: # We want uniq entries
    query.append(record)

def add_nested_value(dic, keys, value, multivalue = False):
    nested_dic = dic
    for key in keys[:-1]:
        if key not in nested_dic:
            nested_dic[key] = {}
        nested_dic = nested_dic[key]
    if not multivalue:
      nested_dic[keys[-1]] = value
    else:
      add_record(nested_dic, keys[-1], value)

# Update
########

def update_innermost(dict1: dict, dict2: dict) -> dict:
    """Recursively updates dict1 with ANY values from dict2 (when the SAME KEY in dict1 is NOT found), 
    OR ONLY with NON-DICT values from dict2 (when the same key in dict1 is found ). Modify dict1 inplace.
    
    Args:
        dict1 (dict): The dictionary to be updated.
        dict2 (dict): The dictionary with values to update dict1.
    Returns:
        None: The function modifies dict1 in place.
    """
    for key, value in dict2.items():
        if isinstance(value, dict) and key in dict1:
            update_innermost(dict1[key], value)
        else:
            dict1[key] = value

# Keys - operation
####################################

def transform_keys(hash, function):
  new_hash = {}
  for key, val in hash.items():
    new_key = function(key)
    new_hash[new_key] = val
  return new_hash

def invert_hash(h):
    new_h = {}
    for k, vals in h.items():
        for v in vals:
            query = new_h.get(v)
            if query == None:
                new_h[v] = [k]
            else:
                query.append(k)
    return new_h

def invert_nested_hash(h):
  new_h = {}
  for k1, vals1 in h.items():
    for k2, vals2 in vals1.items(): 
      query = new_h.get(k2)
      if query == None:
        new_h[k2] = {k1 : vals2}
      else:
        query[k1] = vals2
  return new_h

def remove_nested_entries(nested_hash, func):
  empty_root_ids = []
  for root_id, entries in nested_hash.items():
    delete_entries = []
    for k, v in entries.items():
      if not func(k,v): delete_entries.append(k)
    if len(delete_entries) == len(entries):
      empty_root_ids.append(root_id)
    else:
      for k in delete_entries: entries.pop(k)
  for k in empty_root_ids: nested_hash.pop(k)

####################################
# ARRAY (Matrix) OPERATIONS
####################################

# I/O operations
####################################

def write_lst_files(x_axis_names=None, x_axis_file=None, y_axis_names=None, y_axis_file=None):
    if x_axis_names != None:
      with open(x_axis_file, 'w') as f:
          f.write("\n".join(x_axis_names))
    if y_axis_names != None:
      with open(y_axis_file, 'w') as f:
          f.write("\n".join(y_axis_names))

def save(matrix, matrix_filename, 
  x_axis_names=None, x_axis_file=None, 
  y_axis_names=None, y_axis_file=None, 
  format_type="bin", symm = False, input_matrix_type = "dense", output_matrix_format="dense"):
  if format_type == "bin":
    write_lst_files(x_axis_names, x_axis_file, y_axis_names, y_axis_file)
    if input_matrix_type != output_matrix_format:
        matrix = transform_matrix_format(matrix, output_matrix_format)
  elif input_matrix_type != "dense":
     matrix = transform_matrix_format(matrix, "dense")

  if format_type == "bin":
      if output_matrix_format == "dense":
        np.save(matrix_filename, matrix)
      else:
        save_npz(f"{matrix_filename}.npz", matrix)
  elif format_type == "matrix":
    np.savetxt(matrix_filename, matrix, delimiter='\t')
  elif format_type == "pair":
    pairs = matrix2relations(matrix, x_axis_names, y_axis_names, symm = False)
    with open(matrix_filename, "w") as f:
        for pair in pairs:
            pair[2] = str(pair[2])
            f.write("\t".join(pair) + "\n")

def get_input_source(input_file, format_type):
  if input_file == '-':
      if format_type == "bin":
          data = sys.stdin.buffer.read()
          source = BytesIO(data)
      else:
          source = sys.stdin
  else:
      if format_type == "bin":
          source = open(input_file, 'rb')
      else:
          source = open(input_file)
  return source 

# METHODS FOR TEXT2BINARY
#########################

# def load_matrix_file(source, splitChar = "\t", input_matrix_type="dense"):
#     matrix = None
#     counter = 0
#     col_names = []
#     row_names = []
#     for line in source:
#         line = line.strip()
#         row = [float(c) for c in line.split(splitChar)]
#         if matrix is None:
#             if input_matrix_type == "dense":
#               matrix = np.zeros((len(row), len(row)))
#             elif 
#         for i, val in enumerate(row):
#             matrix[counter, i] = val    
#         counter += 1

#     return matrix



def load_matrix_file(source, splitChar="\t", init_matrix_format="dense"):
    from scipy.sparse import (
      bsr_matrix, coo_matrix, csc_matrix, csr_matrix,
      dia_matrix, dok_matrix, lil_matrix, load_npz
    )
  
    rows = []

    for line in source:
        line = line.strip()
        if not line:
            continue
        row = [float(c) for c in line.split(splitChar)]
        rows.append(row)

    num_rows = len(rows)
    num_cols = len(rows[0]) if rows else 0

    if init_matrix_format in ("dense", "dok", "lil"):
      if init_matrix_format == "dense":
        matrix = np.array(rows)
      elif init_matrix_format == "dok":
        matrix = dok_matrix(rows)
      elif init_matrix_format == "lil":
        matrix = lil_matrix(rows)
    else:
        data, row_idx, col_idx = [], [], []
        for r, row in enumerate(rows):
            for c, val in enumerate(row):
                if val != 0:
                    data.append(val)
                    row_idx.append(r)
                    col_idx.append(c)

        sparse_constructor = {
            "csr": sparse.csr_matrix,
            "csc": sparse.csc_matrix,
            "coo": sparse.coo_matrix,
            "bsr": sparse.bsr_matrix#,
            #"lil": sparse.lil_matrix
        }.get(init_matrix_format)

        if sparse_constructor is None:
            raise ValueError(f"Unsupported sparse matrix type: {init_matrix_format}")

        matrix = sparse_constructor((data, (row_idx, col_idx)), shape=(num_rows, num_cols))

    return matrix

def load_pair_file(source): 
    pairs = []
    for line in source:
        line = line.strip().split("\t")
        if len(line) == 3:
            node_a, node_b, weight = line
            weight = float(weight) 
        else:
            node_a, node_b = line
            weight = 1.0
        pairs.append([node_a, node_b, weight])
    return pairs

def load(matrix_filename, 
          x_axis_file=None, y_axis_file=None, 
          x_idx_file=None, y_idx_file=None,
          format_type="bin", symm=False, 
          init_matrix_format = "dense", output_matrix_format="dense"):
  x_axis = None
  y_axis = None

  source = get_input_source(matrix_filename, format_type)
  if x_axis_file != None:
    with open(x_axis_file, 'r') as f: x_axis = list(map(lambda x: x.rstrip(), f.readlines()))
  if y_axis_file != None:
    with open(y_axis_file, 'r') as f: y_axis = list(map(lambda x: x.rstrip(), f.readlines()))

  x_idxs = None 
  y_idxs = None
  if x_idx_file != None:
    x_idxs = load_name_index(x_idx_file)
  if y_idx_file != None:
    y_idxs = load_name_index(y_idx_file)

  if format_type == 'bin':
    extension = os.path.splitext(matrix_filename)[1][-3:]
    if extension == "npz":
      matrix = load_npz(matrix_filename)
    elif extension == "npy":
      matrix = np.load(matrix_filename)
  elif format_type == 'matrix':
    matrix = load_matrix_file(source, init_matrix_format=init_matrix_format)
  elif format_type == 'pair':
    pairs = load_pair_file(source)
    matrix, x_axis, y_axis = pairs2matrix(pairs, symm=symm, a_index = x_idxs, b_index = y_idxs, init_matrix_format=init_matrix_format)


  if matrix_filename == "-": source.close()
  
  if init_matrix_format != output_matrix_format:
    matrix = transform_matrix_format(matrix, output_matrix_format)

  return matrix, x_axis, y_axis


def transform_matrix_format(input_matrix, output_matrix_format):
  if isinstance(input_matrix, np.ndarray):
    input_matrix = sparse.csr_matrix(input_matrix)
  if output_matrix_format == "bsr":
    matrix = input_matrix.tobsr()
  elif output_matrix_format == "coo":
    matrix = input_matrix.tocoo()
  elif output_matrix_format == "csc":
    matrix = input_matrix.tocsc()
  elif output_matrix_format == "csr":
    matrix = input_matrix.tocsr()
  elif output_matrix_format == "dia":
    matrix = input_matrix.todia()
  elif output_matrix_format == "dok":
    matrix = input_matrix.todok()
  elif output_matrix_format == "lil":
    matrix = input_matrix.tolil()
  elif output_matrix_format == "dense":
    matrix = input_matrix.todense()
  return matrix

# Normalization
####################################

def normalize_matrix(matrix, method="rows_cols"):
  if method == "rows_cols":
    normalized_matrix = row_col_normalization(matrix)
  elif method == "min_max":
    normalized_matrix = min_max_normalization_matrix(matrix)
  elif method == "cosine":
    normalized_matrix = cosine_normalization(matrix)
  else:
    raise ValueError("Invalid normalization method specified.")
  return normalized_matrix

def min_max_normalization_matrix(matrix):
  x_min = np.min(matrix)
  x_max = np.max(matrix)
  normalized_matrix = (matrix - x_min)/(x_max - x_min)
  return normalized_matrix

def z_normalization_matrix(matrix):
  sd = np.std(matrix)
  mean = np.mean(matrix)
  normalized_matrix = (matrix - mean)/sd
  return normalized_matrix

def row_col_normalization(matrix):
  # Classic degree normalization in adjacencies and laplacian matrix.
  D = np.zeros(matrix.shape)
  np.fill_diagonal(D, matrix.sum(axis=1))
  D_inv_sqrt = np.linalg.inv(np.sqrt(D))
  normalized_matrix = np.dot(np.dot(D_inv_sqrt, matrix), D_inv_sqrt)
  return normalized_matrix

def cosine_normalization(matrix):
  # TODO: Need to check when matrix with zeros in diagonal.
  dims = np.shape(matrix)
  normalized_matrix =  np.zeros(dims)
  for i in range(0, dims[0]):
    for j in range(0, dims[1]):
      norm = matrix[i, j]/np.sqrt(matrix[i, i] * matrix[j,j])
      normalized_matrix[i, j] = norm
  return normalized_matrix

# Sim <-> Dis conversion
####################################

def coords2sim(coords, sim = "dotProduct"):
  if sim == "dotProduct":
    sim_mat = coords.dot(coords.T)
  elif sim == "normalizedScaling":
    min_dist = np.min(coords2dis(coords, dist="euclidean"))
    max_dist = np.max(coords2dis(coords, dist="euclidean"))
    sim_mat = 1 - (coords2dis(coords, dist="euclidean") - min_dist) / (max_dist - min_dist)  
  elif isinstance(sim, (float, int)) or sim == "infinity" or sim == "euclidean":
    sim_mat = coords2dis(coords, dist=sim)
    sim_mat = 1 / (sim_mat + 1)
  else:
    raise ValueError("Invalid similarity measure specified.")
  return sim_mat

def coords2dis(coords, dist = "euclidean"):
  if "euclidean": dist = 2
  dist = distance_matrix(coords, coords, p = dist)
  return dist

# Stats
####################################

def get_corr(x = None, y = None, alternative='two-sided', method=None, corr_type= "pearson", rowvar= False):
  if rowvar and y is None:
    # Now, variables are the rows
    x = x.T
  if corr_type == "pearson":
    R, P = pearsonr(x, y, alternative, method)
  elif corr_type == "spearman":
    corr_obj = stats.spearmanr(a=x, b=y, alternative=alternative, nan_policy = "omit")
    R = corr_obj.correlation
    P = corr_obj.pvalue
  else:
    raise ValueError("Invalid correlation type specified.")
  return R, P

def pearsonr(x=None,y=None, alternative='two-sided', method=None): 
  if y is not None:
    corr_obj = stats.pearsonr(x, y, alternative=alternative, method= method)
    R = corr_obj.correlation
    P = corr_obj.pvalue
  else:
    x = x.T
    if alternative == 'two-sided':
        p = lambda t,n: 2 * (1 - stats.t.cdf(abs(t), n - 2))
    elif alternative == 'greater':
        p = lambda t,n: 1 - stats.t.cdf(t, n - 2)
    elif alternative == 'less':
        p = lambda t,n: stats.t.cdf(t, n - 2)
    else:
        raise ValueError("Invalid alternative argument. Valid options are 'two-sided', 'greater', or 'less'.")
    N = np.ones((x.shape[0], x.shape[0]))*x.shape[1] # This should be adopted to na cases?
    R = np.corrcoef(x) 
    T = (R*np.sqrt(N-2))/(np.sqrt(1-R**2))
    P = p(T,N)
  return R, P

def get_stats_from_matrix(matrix): 
  stats = []
  primary_stats = get_primary_stats(matrix)
  #stats << ['Matrix - Symmetric?', matrix.symmetric?]
  stats.append(['Matrix - Dimensions', 'x'.join(map(str, matrix.shape))])
  stats.append(['Matrix - Elements', primary_stats["count"]])
  stats.append(['Matrix - Elements Non Zero', primary_stats["countNonZero"]])
  stats.append(['Matrix - Non Zero Density', primary_stats["countNonZero"]/primary_stats["count"]])
  stats.append(['Weigth - Max', primary_stats["max"]])
  stats.append(['Weigth - Min', primary_stats["min"]])
  stats.append(['Weigth - Average', primary_stats["average"]])
  stats.append(['Weigth - Variance', primary_stats["variance"]])
  stats.append(['Weigth - Standard Deviation', primary_stats["standardDeviation"]])
  stats.append(['Weigth - Q1', primary_stats["q1"]])
  stats.append(['Weigth - Median', primary_stats["median"]])
  stats.append(['Weigth - Q3', primary_stats["q3"]])
  stats.append(['Weigth - Min Non Zero', primary_stats["minNonZero"]])
  stats.append(['Weigth - Average Non Zero', primary_stats["averageNonZero"]])
  stats.append(['Weigth - Variance Non Zero', primary_stats["varianceNonZero"]])
  stats.append(['Weigth - Standard Deviation Non Zero', primary_stats["standardDeviationNonZero"]])
  stats.append(['Weigth - Q1 Non Zero', primary_stats["q1NonZero"]])
  stats.append(['Weigth - Median Non Zero', primary_stats["medianNonZero"]])
  stats.append(['Weigth - Q3 Non Zero', primary_stats["q3NonZero"]])
  connections = get_connection_number(matrix)
  connection_stats = get_primary_stats(connections)
  stats.append(['Node - Elements', connection_stats["count"]])
  stats.append(['Node - Elements Non Zero', connection_stats["countNonZero"]])
  stats.append(['Node - Non Zero Density', connection_stats["countNonZero"]/connection_stats["count"]])
  stats.append(['Edges - Max', connection_stats["max"]])
  stats.append(['Edges - Min', connection_stats["min"]])
  stats.append(['Edges - Average', connection_stats["average"]])
  stats.append(['Edges - Variance', connection_stats["variance"]])
  stats.append(['Edges - Standard Deviation', connection_stats["standardDeviation"]])
  stats.append(['Edges - Q1', connection_stats["q1"]])
  stats.append(['Edges - Median', connection_stats["median"]])
  stats.append(['Edges - Q3', connection_stats["q3"]])
  stats.append(['Edges - Min Non Zero', connection_stats["minNonZero"]])
  stats.append(['Edges - Average Non Zero', connection_stats["averageNonZero"]])
  stats.append(['Edges - Variance Non Zero', connection_stats["varianceNonZero"]])
  stats.append(['Edges - Standard Deviation Non Zero', connection_stats["standardDeviationNonZero"]])
  stats.append(['Edges - Q1 Non Zero', connection_stats["q1NonZero"]])
  stats.append(['Edges - Median Non Zero', connection_stats["medianNonZero"]])
  stats.append(['Edges - Q3 Non Zero', connection_stats["q3NonZero"]])
    
  stats = map(lambda x: [x[0],str(x[1])],stats)
    
  return stats

def get_primary_stats(matrix):
  stats = {}
  max = matrix[0, 0] # Initialize max value
  min = matrix[0, 0] # Initialize min value
  min_non_zero = np.inf # Initialize min value
  values = matrix.flatten()

  stats["count"] = 0
  stats["countNonZero"] = 0 
  stats["sum"] = 0
  for value in values:
    stats["count"] += 1
    stats["countNonZero"] += 1 if value != 0 else 0
    stats["sum"] += value
    max = value if value > max else max
    min = value if value < min else min
    if value != 0 and value < min_non_zero:
      min_non_zero = value
    
  stats["max"] = max
  stats["min"] = min
  stats["minNonZero"] = min_non_zero

  quartile_stats = get_quartiles(values, stats["count"])
  stats.update(quartile_stats)
  non_zero_values = [v for v in values if v != 0]
  quartile_stats_non_zero = get_quartiles(non_zero_values, stats["countNonZero"])
  stats.update(transform_keys(quartile_stats_non_zero, lambda x: x + "NonZero"))
  get_composed_stats(stats, values)
  return stats

def get_connection_number(matrix, count_diagonal=False):
  connections = matrix != 0
  if not count_diagonal: np.fill_diagonal(connections,0)
  connections = connections.sum(1)
  connections = connections[np.newaxis, :]
  return connections

def get_quartiles(values, n_items):
  stats = {}
  stats['q1'] = np.percentile(values,25)
  stats['median'] = np.percentile(values,50)
  stats['q3'] = np.percentile(values,75)
  return stats
  
def get_composed_stats(stats, values):
  average = stats["sum"]/stats["count"]
  average_non_zero = stats["sum"]/stats["countNonZero"]
  stats["average"] = average
  stats["averageNonZero"] = average_non_zero
  stats["sumDevs"] = 0
  stats["sumDevsNonZero"] = 0
  for value in values:
    stats["sumDevs"] += (value - average) ** 2
    stats["sumDevsNonZero"] += (value - average_non_zero) ** 2 if value != 0 else 0
  stats["variance"] = stats["sumDevs"]/stats["count"]
  stats["varianceNonZero"] = stats["sumDevsNonZero"]/stats["countNonZero"]
  stats["standardDeviation"] = stats["variance"] ** 0.5
  stats["standardDeviationNonZero"] = stats["varianceNonZero"] ** 0.5


# Filtering
####################################

def binarize_mat(matrix):
  matrix = matrix > 0
  matrix = matrix.astype(float)
  return matrix
  
def filter_cutoff_mat(matrix, cutoff):
  filtered_mat = matrix >= cutoff
  mat_result = matrix * filtered_mat
  return mat_result

def percentile_filter(matrix, percentile_threshold=90):
  # Note: Percentiles is applied to rows 
  # Could be interesting to expand applications to cols and row_cols
  percentiles = np.percentile(matrix, percentile_threshold, axis=1) # TODO: Check is this is valid for non-square matrix
  result_mat = matrix >= percentiles[:, np.newaxis]
  new_adj = result_mat.transpose()*result_mat
  matrix[~new_adj] = 0
  return matrix

def remove_zero_lines(matrix, rowIds, colIds, symmetric=False):
  # TODO: semgent this in two functions for rows and columns.
  rows2delete = np.all(matrix == 0, axis=1)
  if symmetric:
      cols2delete = rows2delete
  else:
      cols2delete = np.all(matrix == 0, axis=0)

  matrix = matrix[~rows2delete]
  matrix = matrix[:,~cols2delete]
  if rowIds is not None: rowIds = [rowIds[i] for i in range(0,len(rowIds)) if not rows2delete[i]]
  if colIds is not None: colIds = [colIds[i] for i in range(0,len(colIds)) if not cols2delete[i]]

  return matrix, rowIds, colIds


# Clustering
####################################

def get_cluster_by_attributes(matrix, row_names, cluster_method, cluster_kwargs={}):

  if cluster_method == "HDBSCAN":
    from sklearn.cluster import HDBSCAN

    cluster_args_default = {"min_cluster_size": 10,
                    "min_samples": 10, 
                    "metric": "cosine", 
                    "cluster_selection_method": "leaf"}
    cluster_args_default.update(cluster_kwargs)
    hdb = HDBSCAN(**cluster_kwargs)
    hdb.fit(matrix)

    group2sample = {}
    for idx, sample in enumerate(row_names): 
      label = hdb.labels_[idx]
      if label == -1: continue
      if group2sample.get(label):
        group2sample[label].append(sample)
      else:
        group2sample[label] = [sample]
  else:
    raise ValueError(f"Clustering method '{cluster_method}' is not implemented yet.")

  return group2sample

def get_hc_clusters(matrix, dist = 'euclidean', method = 'single', identify_clusters = 'cut_tree', height = None, n_clusters= None, item_list = None, cl_size_factor=1 ):
  from scipy.spatial.distance import pdist, squareform

  cls_objects = {}
  if dist == 'custom': # User gives a custom distance matrix so we only tranfor to pdist output format
    np.fill_diagonal(matrix, 0)
    dist_vector = squareform(matrix)
  else:
    dist_vector = pdist(matrix, metric=dist)
  cls_objects['dist_vector'] = dist_vector
  hc_clust = linkage(dist_vector, method)
  cls_objects['link'] = hc_clust
  if identify_clusters == 'cut_tree':
    clusters = cut_tree(hc_clust, height=height, n_clusters = n_clusters )
  elif identify_clusters == 'max_avg': # use only with matrix distances between 0 - 1
    if dist == 'custom':
      dist_matrix = matrix
    else:
      dist_matrix = squareform(dist_vector)
    clusters = get_cls_max_avg(hc_clust, dist_matrix, cl_size_factor=cl_size_factor)
  else:
    raise ValueError("Invalid cluster identifier metric specified.")
  cls_objects['cls'] = clusters # Give raw clustering object
  tag_singletons(clusters)
  if item_list != None: clusters = tag_items_in_clusters(clusters, item_list)  
  return clusters, cls_objects

def get_cls_max_avg(hc_clust, matrix, cl_size_factor=1):
  y_len, x_len = matrix.shape
  tree = {} # clust id : [member_ids]
  cl_avg_values = {} #clust_id : [ cluster_partition_density ]

  counter = x_len # this works as cluster id. This is used by the linkage method to tag the intermediate clusters: https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html
  last_dist = None
  last_cluster_pool = []
  max_avg_value = 10000000000
  max_cluster_ids = []
  max_avg_val_tree = {}
  for a_id, b_id, dist, n_members in hc_clust:
    dist = round(dist, 5) # To make equal similar distances that differs in very low values
    if last_dist != None and dist != last_dist: # We could have several clusters at the same dist, so we group the merge events to clculate the partition density
      avg_value = get_avg_value4tree_cut(last_cluster_pool, cl_avg_values, tree, x_len, cl_size_factor=cl_size_factor)
      if avg_value < max_avg_value: # check the best partition density
        max_avg_value = avg_value
        max_cluster_ids = last_cluster_pool
        max_avg_val_tree = copy.copy(tree)
    a_id = int(a_id) # Linkage method returns member ids as float instead of int
    b_id = int(b_id)
    member_list = get_member_list(counter, a_id, b_id, x_len, tree) # members that we merge to build the new agglomerative cluster
    cl_avg_values[counter] = get_cluster_avg_value(member_list, matrix)
    last_cluster_pool = [ cl_id for cl_id in last_cluster_pool if cl_id not in [a_id, b_id] ] # update clusters removin merged cl ids and adding the new cluters ids
    last_cluster_pool.append(counter)
    last_dist = dist
    counter += 1
      
  avg_value = get_avg_value4tree_cut(last_cluster_pool, cl_avg_values, tree, x_len, cl_size_factor=cl_size_factor) # update clusters removin merged cl ids and adding the new cluters ids
  if avg_value < max_avg_value: # check the best partition density on the last distance that not was checked
    max_avg_value = avg_value
    max_cluster_ids = last_cluster_pool
    max_avg_val_tree = copy.copy(tree)
  final_clusters = np.repeat([[None]], x_len, axis=0)
  for cluster_id in max_cluster_ids:
    members = max_avg_val_tree[cluster_id]
    for m in members: final_clusters[m] = [cluster_id]
  return final_clusters

def get_avg_value4tree_cut(last_cluster_pool, cl_avg_values, tree , n_elements, cl_size_factor=1):
  avg_value = np.median([ cl_avg_values[cl_id] for cl_id in last_cluster_pool])
  total = sum([ len(tree[cl_id]) for cl_id in last_cluster_pool])
  avg_value = (1 - total/(n_elements+1)/cl_size_factor) * (1 + avg_value)
  # avg_value = sum([ cl_avg_values[cl_id] * len(tree[cl_id]) for cl_id in last_cluster_pool])/total
  return avg_value

def get_member_list(cluster_id, a_id, b_id, n_records, tree):
  member_list = []
  add_cluster_members(member_list, a_id, n_records, tree) # get cluster members from previous a cluster
  add_cluster_members(member_list, b_id, n_records, tree) # get cluster members from previous b cluster
  tree[cluster_id] = member_list
  return member_list

def add_cluster_members(cluster, member_id, n_records, tree):
  if member_id < n_records: # check if member_id is a cluster with only one member that is a original record. That id is less than n_records is the criteria described in https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html
    cluster.append(member_id)
  else: # The id represents the merge of two previous clusters. We obtain the member list from the tree and remove it to merge it in the new cluster
    cluster.extend(tree.pop(member_id))

def get_cluster_avg_value(member_list, matrix):
  members = copy.copy(member_list)
  values = []
  while len(members) > 1:
    ref = members.pop()
    values.extend([matrix[ref, m] for m in members])
  return np.mean(values)

def tag_singletons(clusters):
  counter = 0
  for i in range(0, len(clusters)):
    if clusters[i][0] is None:
      clusters[i] = [i]
      counter += 1

def tag_items_in_clusters(data, item_list):
  clustering_data = []
  for i, cluster_id in enumerate(data): clustering_data.append([item_list[i], cluster_id[0]])
  clusters = {}
  for item_id, cluster_id in clustering_data:
    query = clusters.get(cluster_id)
    if query == None:
        clusters[cluster_id] = [item_id]
    else:
        query.append(item_id)
  return clusters  

####################################
# Explicit conversions
####################################

# General conversor #

def transform2obj(obj, inFormat=None, outFormat=None, rowIds= None, colIds= None, **kwargs): 
  if outFormat == 'pair':
    if inFormat == 'matrix': 
        obj = matrix2pairs(obj, rowIds=rowIds, colIds=colIds)
    if inFormat == 'nested_pairs':
        obj = nested_pairs2pairs(obj)
  elif outFormat == 'matrix':
    if inFormat == 'pair': 
        obj, rowIds, colIds = pairs2matrix(obj)
    if inFormat == 'nested_pairs':
        obj, rowIds, colIds = pairs2matrix(nested_pairs2pairs(obj)) # Talk with PSZ about the trade-off combinatorial vs optimization
  elif outFormat == "nested_pairs":
    if inFormat == "pair":
        obj = list2dic(obj)
  elif outFormat == "lists":
    if inFormat == "dataframe":
      obj = dataframe2lists(obj, kwargs.get("rownames_name"))
  else:
    raise ValueError("Invalid conversion specified.")
  return obj, rowIds, colIds

# DataFrame -> List of Lists
###################################

def dataframe2lists(df, rownames_name):
  rownames_name_val = rownames_name or df.index.name or "samples"
  columns = [rownames_name_val] + [i for i in df.columns]
  rows = [[i for i in row] for row in df.itertuples()]
  table = [columns] + rows
  return table

# List -> Dic
####################################

def flatlist2dic(l):
  return {el:True for el in l}

# List -> dic
#############

def list2dic(pairs):
  dic_from_list = {}
  for pair in pairs: add_nested_value(dic_from_list, keys=pair[:-1],value=pair[-1])
  return dic_from_list

# Dic -> Matrix(array)
####################################

def to_bmatrix(dictio):
  x_names_indx = get_hash_values_idx(dictio)
  y_names = list(dictio.keys())
  x_names = list(x_names_indx.keys())
   # row (y), cols (x)
  matrix = np.zeros((len(dictio), len(x_names)))
  i = 0
  for id, items in dictio.items():
    for item_id in items: matrix[i, x_names_indx[item_id]] = 1
    i += 1
  return matrix, y_names, x_names

def to_wmatrix(dictio, squared = True, symm = True):
  if squared:
    matrix, element_names = to_wmatrix_squared(dictio, symm=symm)
    return matrix, element_names
  else:
    matrix, y_names, x_names = to_wmatrix_rectangular(dictio)
    return matrix, y_names, x_names

def to_wmatrix_squared(dictio, symm = True):
  element_names = {}
  for elementA, relations in dictio.items():
    if element_names.get(elementA) == None:
      element_names[elementA] = True
    for elementB in relations.keys():
      if element_names.get(elementB) == None: 
        element_names[elementB] = True
  
  element_names = list(element_names.keys())
  matrix = np.zeros((len(element_names), len(element_names)))
  i = 0
  for elementA in element_names:
    relations = dictio.get(elementA)
    if relations != None:
      for j, elementB in enumerate(element_names):
        if elementA != elementB:
          query = relations.get(elementB)
          if query != None:
            matrix[i, j] = query
            if symm: # TODO: PSZ, lo q se hace aqui no me cuadra
              matrix[j, i] = query
    i += 1
  return matrix, element_names

def to_wmatrix_rectangular(dictio):
  y_names = list(dictio.keys())
  x_names = list(get_hash_values_idx(dictio).keys())
  matrix = np.zeros((len(y_names), len(x_names)))
  i = 0
  for elementA, relations in dictio.items():
    for j, elementB in enumerate(x_names):
        query = relations.get(elementB)
        if query != None:
          matrix[i, j] = query
    i += 1
  return matrix, y_names, x_names

# Dic -> List
####################################

def nested_pairs2pairs(nested_dic_pairs): 
  pairs = []
  for item_a, dat in nested_dic_pairs.items(): 
    for item_b, val in dat.items(): pairs.append([item_a, item_b, val])
  return pairs

# List -> Matrix(Array)
####################################

def load_name_index(path):
  index = {}
  with open(path) as f:
    for i, line in enumerate(f): index[line.rstrip()] = i
  return index

def pairs2matrix(pairs, symm=True, a_index=None, b_index=None, init_matrix_format="dense"):
    count_A = 0
    index_A = {}
    if a_index: index_A = a_index
    count_B = 0
    index_B = {}
    if b_index: index_B = b_index

    # Indexing
    if symm:
        for elementA, elementB, val in pairs:
            count_A = update_index(index_A, elementA, count_A)
            count_A = update_index(index_A, elementB, count_A)
        index_B = index_A
    else:
        for elementA, elementB, val in pairs:
            count_A = update_index(index_A, elementA, count_A)
            count_B = update_index(index_B, elementB, count_B)

    elementA_names = list(index_A.keys())
    elementB_names = list(index_B.keys())
    num_rows = len(elementA_names)
    num_cols = len(elementB_names)

    # Create matrix
    if init_matrix_format in ("dense", "dok", "lil"):
        if init_matrix_format == "dense":
          matrix = np.zeros((num_rows, num_cols))
        elif init_matrix_format == "dok":
          matrix = sparse.dok_matrix((num_rows, num_cols))
        elif init_matrix_format == "lil":
          matrix = sparse.lil_matrix((num_rows, num_cols))
        for elementA, elementB, val in pairs:
            i = index_A[elementA]
            j = index_B[elementB]
            matrix[i, j] = val
            if symm:
                matrix[j, i] = val
    else:
        row, col, data = [], [], []
        for elementA, elementB, val in pairs:
            i = index_A[elementA]
            j = index_B[elementB]
            row.append(i)
            col.append(j)
            data.append(val)
            if symm and i != j:
                row.append(j)
                col.append(i)
                data.append(val)

        # Sparse matrix creation
        sparse_constructor = {
            "csr": sparse.csr_matrix,
            "csc": sparse.csc_matrix,
            "coo": sparse.coo_matrix,
            "bsr": sparse.bsr_matrix#,
            #"dia": sparse.dia_matrix,
        }.get(init_matrix_format)

        if sparse_constructor is None:
            raise ValueError(f"Unsupported sparse format: {init_matrix_format}")

        matrix = sparse_constructor((data, (row, col)), shape=(num_rows, num_cols))

    return matrix, elementA_names, elementB_names

def update_index(index, element, count): 
  # TODO: Convert index to default dictionary with 0 value
  if index.get(element) is None:
    index[element] = count
    count += 1
  return count

# Matrix(Array) -> List
####################################
# TODO: Talk with PSZ: The next three methods are redundant, 
# but i dont see an efficient manner of making one generic function (Fred)

def matrix2pairs(matrix, rowIds, colIds, symm = False): 
  relations = []
  if symm:
    for rowPos, rowId in enumerate(rowIds):
      for colPos, colId in enumerate(colIds[rowPos:]):
        colPos += rowPos
        relations.append([rowId, colId, matrix[rowPos, colPos]])
  else:
    for rowPos, rowId in enumerate(rowIds):
      for colPos, colId in enumerate(colIds):
        relations.append([rowId, colId, matrix[rowPos, colPos]])
  return relations

def matrixes2pairs(matrixes, rowIds, colIds, symm = False): 
  # When there are multiple matrix with the same rows and cols.
  relations = []
  if symm:
    for rowPos, rowId in enumerate(rowIds):
      for colPos, colId in enumerate(colIds[rowPos:]):
        colPos += rowPos
        associationValues = [matrix[rowPos, colPos] for matrix in matrixes]
        relations.append([rowId, colId, *associationValues])
  else:
    for rowPos, rowId in enumerate(rowIds):
      for colPos, colId in enumerate(colIds):
        associationValues = [matrix[rowPos, colPos] for matrix in matrixes]
        relations.append([rowId, colId, *associationValues])
  return relations

def matrix2relations(finalMatrix, rowIds, colIds, symm = False): 
  relations = []
  if symm:
    for rowPos, rowId in enumerate(rowIds):
      for colPos, colId in enumerate(colIds[rowPos:]):
        colPos += rowPos
        associationValue = finalMatrix[rowPos, colPos]
        if associationValue > 0: relations.append([rowId, colId, associationValue])
  else:
    for rowPos, rowId in enumerate(rowIds):
      for colPos, colId in enumerate(colIds):
        associationValue = finalMatrix[rowPos, colPos]
        if associationValue > 0: relations.append([rowId, colId, associationValue])
  return relations
